# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import collections.abc as cabc

from .nodes.dict import ConfigDict
from .eval_context import EvalContext
from .namespace import namespace, NamespaceableMeta
from .utils import Bunch
from .nodes.required import RequiredNode


class Config(Bunch, metaclass=NamespaceableMeta):
    ''' A class representing parsed and evaluated config dictionary.
    '''
    def __init__(self, config_dict=None):
        ''' Arguments:
                config_dict : `dict` or unevaluated `awesomeyaml.nodes.ConfigDict`, `None`
                    represents an empty dict.
        '''
        if config_dict is not None and not isinstance(config_dict, dict):
            raise ValueError('dict or None expected')

        if not isinstance(config_dict, ConfigDict):
            config_dict = ConfigDict(config_dict)

        if config_dict:
            Config.check_missing(config_dict)
            self._source = config_dict
            pre_evaluate = copy.deepcopy(config_dict)
            eval_ctx = EvalContext(pre_evaluate)
            evaluated = eval_ctx.evaluate()
        else:
            evaluated = {}

        super().__init__(evaluated)

    @classmethod
    def build(cls, *sources, raw_yaml=None, filename=None):
        ''' Builds a config from the provided yaml sources and evaluates it, returning `awesomeyaml.Config` object.

            Arguments:
                *sources : a list of yaml sources - that can include file-like objects, filenames and strings of yaml
                raw_yaml : 
        '''
        from .builder import Builder
        b = Builder()
        b.add_multiple_sources(*sources, raw_yaml=raw_yaml, filename=filename)
        return Config(b.build())

    @classmethod
    def process_cmdline(cls, args, filename_lookup_fn=None):
        yamls = []
        filenames = []
        raw_yamls = []

        def append(idx, option):
            if option.startswith('{') and option.endswith('}'):
                yaml = option
                filename = f'<Commandline argument #{idx}>'
                raw_yaml = True
            elif '=' in option:
                key, value = option.split('=', maxsplit=1)
                yaml = ''
                ind = 0
                for part in key.split('.'):
                    if ind:
                        yaml += '\n'
                        yaml += '    ' * ind

                    yaml += part + ': '
                    ind += 1

                yaml += value
                filename = f'<Commandline argument #{idx}>'
                raw_yaml = True
            else:
                yaml = option
                if filename_lookup_fn is not None:
                    yaml = filename_lookup_fn(yaml)
                filename = yaml
                raw_yaml = False

            yamls.append(yaml)
            filenames.append(filename)
            raw_yamls.append(raw_yaml)

        for i, src in enumerate(args):
            append(i, src)

        return yamls, filenames, raw_yamls

    @classmethod
    def build_from_cmdline(cls, *sources, filename_lookup_fn=None):
        yamls, filenames, raw_yamls = cls.process_cmdline(sources, filename_lookup_fn=filename_lookup_fn)
        return cls.build(*yamls, raw_yaml=raw_yamls, filename=filenames)

    @staticmethod
    def check_missing(cfg):
        missing = []
        for path, node in cfg.ayns.nodes_with_paths():
            if isinstance(node, RequiredNode):
                missing.append(repr(path))

        if missing:
            raise ValueError('The following required nodes have not been set:\n    ' + '\n    '.join(missing))

    @namespace('ayns')
    @property
    def source(self):
        ''' The source `awesomeyaml.nodes.ConfigDict` which was evaluated to construct
            this `awesomeyaml.Config`.
        '''
        return self._source

    @namespace('ayns')
    def pprint(self, ind=2, init_level=0):
        import io
        import functools

        try:
            import torch.nn as nn
        except:
            nn = None

        ind = ' '*ind
        ret = '' #ind*init_level
        stack = [[self, init_level, None]]
        emitted_recently = False
        offset = 0
        while stack:
            frame = stack[-1]
            obj, level, i = frame
            to_emit = obj
            finished = False
            emit = False
            if i is None:
                if isinstance(obj, functools.partial):
                    args = obj.keywords.copy()
                    if obj.args:
                        args.update({ idx: arg for idx, arg in enumerate(obj.args) })

                    to_emit = obj.func
                    obj = args
                    frame[0] = obj
                    emit = True
                elif nn is not None and isinstance(obj, nn.Module):
                    obj = f'<Neural network of type \'{type(obj).__module__}.{type(obj).__qualname__}\' with {sum(p.numel() for p in obj.parameters())} parameters and {sum(1 for _ in obj.modules())} submodules>'
                    to_emit = obj


                if not isinstance(obj, (cabc.Sequence, cabc.Mapping)) or isinstance(obj, (str, bytes, io.IOBase)):
                    finished = True
                    emit = True
                    if isinstance(obj, str) and '\n' in obj:
                        str_ind = ind*(level+1)+'    '
                        line_ind = str_ind + '    '
                        to_emit = '\n' + str_ind + "''' " + f'\n{line_ind}'.join(obj.rstrip().split('\n')) + '\n' + str_ind + "'''\n"
                else:
                    try:
                        i = iter(obj)
                    except TypeError:
                        finished = True
                        emit = True
                    else:
                        if not obj:
                            finished = True
                            emit = True
                            i = None
                        elif Config._pprint_is_simple_list(obj):
                            offset_fix = 0
                            if len(stack) > 1 and isinstance(stack[-2][0], dict):
                                offset_fix = 1 # account for the extra space that is added in the `elif emit:` block below
                            to_emit = Config._pprint_format_simple_list(obj, ind*level, width_limit=160, offset=offset+offset_fix)
                            finished = True
                            emit = True
                            i = None

                if not finished:
                    frame[2] = i
                    if len(stack) > 1 and isinstance(stack[-2][0], dict): # if a dict contains a collection, add newline after key
                        if not emit:
                            ret += '\n'
                            offset = 0
                        else: # the same as below, but special case when `not finished and emit`
                            ret += ' '
                            offset += 1
                elif emit:
                    if len(stack) > 1 and isinstance(stack[-2][0], dict): # if a dict contains an empty collection or scalar value, add space after key
                        ret += ' '
                        offset += 1

            if emit:
                try:
                    ret += str(to_emit)
                except ValueError:
                    ret += repr(to_emit)
                ret += '\n'
                offset = 0
                emitted_recently = True

            if not finished:
                try:
                    child = next(i)
                except StopIteration:
                    finished = True
                else:
                    if isinstance(obj, dict):
                        ret += ind*level
                        ret += str(child)
                        ret += ':'
                        offset += len(ind) * level + len(str(child)) + 1
                        child = obj[child]
                    else:
                        ret += ind*level
                        ret += '- '
                        offset += len(ind) * level + 2

                    stack.append([child, level+1, None])

            if finished:
                stack.pop()
                # we finished a non-empty interable, add extra newline
                if i is not None and emitted_recently:
                    ret += '\n'
                    offset = 0
                    emitted_recently = False

        return ret

    @staticmethod
    def _pprint_is_simple_list(obj):
        if isinstance(obj, cabc.Sequence) and not isinstance(obj, (str, bytes)):
            if all(isinstance(e, (int, float)) or (isinstance(e, str) and len(e) < 25 and '\n' not in e) for e in obj):
                return True
        return False

    @staticmethod
    def _pprint_format_simple_list(obj, ind, width_limit=160, offset=0):
        width_limit += max(0, max(len(ind), offset) - width_limit + 40) # make sure we have at least 40 characters per line, violate width limit if necessary
        ret = ''
        line = ''
        sep = '['
        size = len(obj)
        for i, e in enumerate(obj):
            to_add = sep + repr(e)
            if i+1 < size:
                to_add += ','
            else:
                to_add += ']'

            if len(line) + len(to_add) + offset >= width_limit:
                ret += line + '\n'
                line = ind + to_add
                offset = 0
            else:
                line += to_add
                sep = ' '

        ret += line
        return ret



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

version = '1.0.0'
repo = 'unknown'
commit = 'unknown'
has_repo = False

try:
    import git
    from pathlib import Path

    try:
        r = git.Repo(Path(__file__).parents[1])
        has_repo = True

        if not r.remotes:
            repo = 'local'
        else:
            repo = r.remotes.origin.url

        commit = r.head.commit.hexsha
        if r.is_dirty():
            commit += ' (dirty)'
    except git.InvalidGitRepositoryError:
        raise ImportError()
except ImportError:
    pass

try:
    from . import _dist_info as info
    assert not has_repo, '_dist_info should not exist when repo is in place'
    assert version == info.version
    repo = info.repo
    commit = info.commit
except (ImportError, SystemError):
    pass

__all__ = ['version', 'repo', 'commit', 'has_repo']



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Namespace():
    def __init__(self, cls, name, names):
        self._cls = cls
        self._name = name
        self._names = names

    def __repr__(self):
        return f'<Namespace {self._cls.__name__}.{self._name} with type {type(self).__name__!r} at 0x{id(self):02x}>'

    def __get__(self, inst, cls=None):
        if inst is None:
            return self
        return Namespace.Bind(self, inst)

    def __set__(self, inst, value):
        raise AttributeError("can't set namespace")

    def __delete__(self, inst):
        raise AttributeError("can't delete namespace")

    def _resolve_endpoint(self, name):
        for cls in self._cls.__mro__:
            namespace = cls.__dict__.get(self._name, None)
            if namespace is None or not isinstance(namespace, Namespace):
                continue

            if name in namespace._names:
                return namespace._names[name]

        raise AttributeError(f'{self!r} does not have attribute {name!r}')

    def __getattr__(self, name):
        endpoint = self._resolve_endpoint(name)
        return endpoint.__get__(None, self._cls)

    def __setattr__(self, name, value):
        if not name.startswith('_'):
            raise AttributeError("Can't set a new public attribute to a namespace")

        return super().__setattr__(name, value)

class BoundNamespace(Namespace):
    def __init__(self, namespace, inst):
        super().__init__(namespace._cls, namespace._name, namespace._names)
        self._orig_namespace = namespace
        self._inst = inst

    def __repr__(self):
        return f'<Bound namespace {self._cls.__name__}.{self._name} with type {type(self._orig_namespace).__name__!r} of {self._inst!r}>'

    def __getattr__(self, name):
        endpoint = self._resolve_endpoint(name)
        return endpoint.__get__(self._inst, type(self._inst))

    def __setattr__(self, name, value):
        if not name.startswith('_'):
            error = False
            try:
                endpoint = self._resolve_endpoint(value)
            except AttributeError:
                error = True

            if not error:
                return endpoint.__set__(self._inst, value)

        return super().__setattr__(name, value)

Namespace.Bind = BoundNamespace


class staticproperty(property):
    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        if fget is not None and not isinstance(fget, staticmethod):
            raise ValueError('fget should be a staticmethod')
        if fset is not None and not isinstance(fset, staticmethod):
            raise ValueError('fset should be a staticmethod')
        if fdel is not None and not isinstance(fdel, staticmethod):
            raise ValueError('fdel should be a staticmethod')
        super().__init__(fget, fset, fdel, doc)

    def __get__(self, inst, cls=None):
        if inst is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        return self.fget.__get__(inst, cls)() # pylint: disable=no-member

    def __set__(self, inst, val):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        return self.fset.__get__(inst)(val) # pylint: disable=no-member

    def __delete__(self, inst):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        return self.fdel.__get__(inst)() # pylint: disable=no-member


class namespaceable_property(property):
    def __init__(self, fget, fset, fdel, doc, namespace_data):
        super().__init__(fget, fset, fdel, doc)
        self._awesomeyaml_namespace = namespace_data

    def getter(self, func):
        return type(self)(func, self.fset, self.fdel, self.__doc__, self._awesomeyaml_namespace)

    def setter(self, func):
        return type(self)(self.fget, func, self.fdel, self.__doc__, self._awesomeyaml_namespace)

    def deleter(self, func):
        return type(self)(self.fget, self.fset, func, self.__doc__, self._awesomeyaml_namespace)


def namespace(namespace_name, namespace_type=Namespace):
    if not isinstance(namespace_name, str) or not namespace_name:
        raise ValueError(f'Namespace name should be a non-empty str, got: {namespace_name}')
    def namespace_decorator(func):
        namespace_data = (namespace_name, namespace_type)
        if isinstance(func, property) and not isinstance(func, namespaceable_property):
            func = namespaceable_property(func.__get__, func.__set__, func.__delete__, func.__doc__, namespace_data)
        else:
            func._awesomeyaml_namespace = namespace_data
        return func
    return namespace_decorator


class NamespaceableMeta(type):
    def __init__(cls, clsname, clsbases, clsnamespace, **kwargs):
        namespaces = {}
        for name, value in clsnamespace.items():
            if isinstance(value, type) and issubclass(value, Namespace):
                if len(value.__bases__) != 1:
                    raise ValueError('Expected exactly one base for an embedded Namespace type')
                namespace_name = name
                namespace_type = value.__bases__[0]
            elif hasattr(value, '_awesomeyaml_namespace'):
                namespace_name, namespace_type = value._awesomeyaml_namespace
            else:
                continue

            if name.startswith('__'):
                raise ValueError(f'Private entities (with names starting with "__") cannot be put in a namespace, when processing namespace attribute for entity: {value}, with name: {name!r}')
            if namespace_name in clsnamespace and not (isinstance(clsnamespace[namespace_name], type) and issubclass(clsnamespace[namespace_name], Namespace)):
                raise ValueError(f'Cannot hide a class member {name!r} under a namespace {namespace_name!r} as the namespace name {namespace_name!r} is already in use by {clsnamespace[namespace_name]}')
            if namespace_name in namespaces and namespaces[namespace_name][None] is not namespace_type:
                raise ValueError(f'Namespace {namespace_name!r} has inconsistent type: currently requested {namespace_type}, previously: {namespaces[namespace_name][None]}')

            namespace_names = namespaces.setdefault(namespace_name, { None: namespace_type })
            if isinstance(value, type):
                for subname, subvalue in value.__dict__.items():
                    namespace_names[subname] = subvalue
            else:
                namespace_names[name] = value

        for namespace_name, namespace_values in namespaces.items():
            namespace_type = namespace_values.pop(None)
            setattr(cls, namespace_name, namespace_type(cls, namespace_name, namespace_values))
            for name, value in namespace_values.items():
                # we need to take care of fields like '__module__'
                if name is not None and name in cls.__dict__ and not name.startswith('__'):
                    delattr(cls, name)


class Namespaceable(metaclass=NamespaceableMeta):
    pass



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .scalar import ConfigScalar
from ..namespace import namespace, staticproperty
from ..utils import import_name


class ImportNode(ConfigScalar(str)):
    ''' Implements ``!import`` tag.

        The import node can be used to embed any python entity within the config.
        Internally, it is treated as a standard string node with the only exception
        being its evaluation which is implemented by the means of :py:func:`awesomeyaml.utils.import_name`.
        Specifically, the evaluated value is defined as::

            return import_name(str(self))

        Please see :py:func:`import_name` function for more details about how name
        lookup is performed.

        Supported syntax::

            !import name

        Merge behaviour:

            The same as standard string node.
    '''
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        return import_name(str(self))

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def tag():
        return '!import'



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import collections.abc as cabc

from .node import ConfigNode
from ..namespace import Namespace, staticproperty


class NodePath(list):
    def __str__(self):
        return ComposedNode.get_str_path(self)

    def __repr__(self):
        return repr(self.__str__())

    def __add__(self, other):
        return NodePath(list.__add__(self, other))


class ComposedNode(ConfigNode):
    _path_component_regex = re.compile(r'''
                # the available options are:
                    (?:^|(?<=\.)) # ...if it is either at the beginning or preceded by a dot (do not capture)
                    ( [a-zA-Z0-9_]+ ) # a textual identifier (captured by the fist capture group)
                | # OR...
                    \[ (-?[0-9]+) \] # a (possibly negative) integer in [] (with second capture group catching the integer inside)
                | # OR...
                    (?<!^) # ... if it's not the first character ...
                    \. # a single dot
                    (?:(?!$)(?!\.)(?!\[)) # ... does not appear at the end and is not followed by either another dot or [ (do not capture)
                ''', re.VERBOSE) # verbose flag enables us to have comments, whitespace (inc. multi-line) etc. for better readability

    @staticmethod
    def split_path(path_str, validate=True):
        matches = ComposedNode._path_component_regex.finditer(path_str)
        if validate:
            _matches_cache = []
            beg = None
            end = None
            for match in matches:
                _matches_cache.append(match)
                if beg is None:
                    beg = match.start()
                # in the first iteration, check is beg (i.e. match.start()) == 0 to make sure we do not have an invalid prefix
                # in every other iteration make sure the new match strictly follow the previous one, if it doesn't that means
                # that there was an invalid part between them (invalid here means not matching our regular expression)
                if match.start() != (end if end is not None else 0):
                    raise ValueError(f'Invalid path: {path_str!r}')
                end = match.end()

            # the path is fine at the beginning and internally doesn't have any "gaps" but has an incorrect suffix
            # (i.e. a part at the end which doesn't match our regular expression)
            if path_str and end != len(path_str):
                raise ValueError(f'Invalid path: {path_str!r}')

            matches = _matches_cache
        
        for match in matches:
            if match.group(1): # name
                yield str(match.group(1))
            elif match.group(2):  # index
                yield int(match.group(2))
            else:
                # sanity check
                assert match.group(0) == '.'

    @staticmethod
    def join_path(path_list):
        ret = ''
        for component in path_list:
            ret = ret + ComposedNode._get_child_accessor(component, ret)
        return ret

    @staticmethod
    def _get_child_accessor(childname, myname=''):
        if isinstance(childname, int):
            return f'[{childname}]'
        return ('.' if str(myname) else '') + str(childname)

    @staticmethod
    def get_list_path(*path, check_types=True):
        if not path:
            return NodePath()
        if len(path) == 1:
            if not isinstance(path[0], int):
                path = path[0]

        if path is None:
            return NodePath()
        if not isinstance(path, cabc.Sequence) or isinstance(path, str):
            # split_path should only return str and ints so we don't need to check for types
            path = ComposedNode.split_path(str(path))
        elif check_types:
            for i, c in enumerate(path):
                # we need to check it because if something is not a string nor an int it's ambiguous which casting should be done
                if not isinstance(c, str) and (not isinstance(c, int) or isinstance(c, bool)):
                    raise ValueError(f'node path should be a list of ints and/or str only, got {type(c)} at position {i}: {c}')

        return NodePath(path)

    @staticmethod
    def get_str_path(*path, check_types=True):
        if not path:
            return ''
        if len(path) == 1:
            path = path[0]

        if path is None:
            return ''
        if isinstance(path, int):
            return ComposedNode._get_child_accessor(path)
        if isinstance(path, cabc.Sequence) and not isinstance(path, str):
            return ComposedNode.join_path(path)

        if check_types and not isinstance(path, str):
            raise ValueError(f'Unexpected type: {type(path)}, expected int, sequence or str')
        return path

    @staticmethod
    def maybe_inherit_flags(src, kwargs):
        if isinstance(src, ConfigNode):
            kwargs.setdefault('idx', src._idx)
            kwargs.setdefault('delete', src._delete)
            kwargs.setdefault('merge_mode', src._merge_mode)
            kwargs.setdefault('metadata', src._metadata)
            kwargs.setdefault('implicit_delete', src._implicit_delete)

    def __init__(self, children, nodes_memo=None, **kwargs):
        super().__init__(**kwargs)
        kwargs.pop('idx', None)
        kwargs.pop('metadata', None)
        if kwargs.pop('delete', None):
            kwargs['implicit_delete'] = True
        nodes_memo = nodes_memo if nodes_memo is not None else {}
        self._children = { name: ConfigNode(child, **kwargs, nodes_memo=nodes_memo) for name, child in children.items() } # pylint: disable=unexpected-keyword-arg

    class ayns(Namespace):
        def set_child(self, name, value):
            value = ConfigNode(value)
            self._children[name] = value
            return value

        def remove_child(self, name):
            child = self._children.pop(name, None)
            return child

        def rename_child(self, old_name, new_name):
            if not self.ayns.has_child(old_name):
                raise ValueError(f'{self!r} does not have a child named: {old_name!r}')
            if self.ayns.has_child(new_name):
                raise ValueError(f'Cannot rename a child named: {old_name!r} to {new_name!r}, the new name is already assigned to another child')

            child = self._children[old_name]
            del self._children[old_name]
            self._children[new_name] = child
            return child

        def get_child(self, name, default=None):
            return self._children.get(name, default)

        def has_child(self, name):
            if '_children' not in self.__dict__:
                return False
            return name in self._children

        @staticmethod
        def _get_node(root, access_fn, query_fn, *path, intermediate=False, names=False, incomplete=None):
            path = ComposedNode.get_list_path(*path)

            ret = []
            def _add(child, name):
                if names:
                    path = []
                    if ret:
                        path = ret[-1][2] + [name]
                    ret.append((child, name, path))
                else:
                    ret.append(child)

            _add(root, None)
            node = root
            found = True
            for component in path:
                if not query_fn(node, component): #not isinstance(node, ComposedNode) or not node.ayns.has_child(component):
                    _add(None, component)
                    found = False
                    break

                #node = node.ayns.get_child(component)
                node = access_fn(node, component)
                _add(node, component)

            if not found and not incomplete:
                if incomplete is None:
                    return None

                raise KeyError(f'{ComposedNode.join_path(path)!r} does not exist within {root!r}')

            return ret if intermediate else ret[-1]

        @staticmethod
        def _remove_node(root, remove_fn, *path):
            nodes = root.ayns.get_node(*path, intermediate=True, names=True, incomplete=None)
            if nodes is None:
                return None
            if len(nodes) == 1:
                raise ValueError(f'Cannot remove self from self')
            assert len(nodes) >= 2
            node, name, _ = nodes[-1]
            assert node is not None
            parent, _, _ = nodes[-2]
            return remove_fn(parent, name)

        def get_node(self, *path, intermediate=False, names=False, incomplete=False):
            ''' Inputs:
                    - `path` either a list of names (str or int) which should be looked up, or a str equivalent to calling ComposedNode.join_path on an analogical list
                    - `intermediate` if True, returned is a list of nodes accessed (including self), in order of accessing, otherwise only the final node is returned (i.e. the last element of the list)
                    - `names` if True, returned are tuples of `(node, name, path)`, where `node` is an object represting a node with name `name` (relative to its parent) and path `path` (list of names, relative to the self),
                        otherwise only the node object is returned.
                        This option applies to both cases: with and without `intermediate` set to True - in both cases either the single value returned, or all the values in the returned list, are either tuples of single objects.
                        If `intermediate` is True, or `path` is empty list/None, `self` can be also returned in which case `name` is None and `path` is an empty list.
                    - `incomplete` if evaluates to True and a node with path `path` cannot be found, returns the first missing node as `None` (if `names` is False) or `(None, name, path)` - in case `intermediate` is also set to True,
                        the last element of the returned list will be the first missing node with the rest of the list populated as usual. If `incomplete` is `None` and the node doesn't exist, `None` is returned (`intermediate` and `names` are ignored in that case),
                        otherwise a `KeyError` is raised.

                Returns:
                    - a node found by following path `path`, optionally together with its name and the full path (if `names` is True), if the node exists,
                    - otherwise if `incomplete` is set to True, a node "found" is the first missing node and is returned as `None` (again, optionally with name and a full path),
                    - regardless of whether a node was found or None was used to indicated a missing node, if `intermediate` is set to True returned is a list of all nodes
                        which were accessed (in order of accessing) before the final node was reached - this includes `self`, also each node can be returned either on its own or as a tuple together with their names and paths

                Raises:
                    `KeyError` if node cannot be found and `incomplete` evaluates to False and is not None
            '''
            def access_fn(node, component):
                return node.ayns.get_child(component)
            def query_fn(node, component):
                if not isinstance(node, ComposedNode):
                    return False
                return node.ayns.has_child(component)

            return ComposedNode.ayns._get_node(self, access_fn, query_fn, *path, intermediate=intermediate, names=names, incomplete=incomplete)

        def get_first_not_missing_node(self, *path, intermediate=False, names=False):
            def _get_node(n):
                return n if not names else n[0]
            nodes = self.ayns.get_node(*path, intermediate=True, names=names, incomplete=True)
            assert len(nodes) >= 2 or nodes[-1] is self
            if _get_node(nodes[-1]) is not None:
                return nodes[-1] if not intermediate else nodes

            nodes.pop()
            assert _get_node(nodes[-1]) is not None
            return nodes[-1] if not intermediate else nodes

        def remove_node(self, *path):
            def remove_fn(node, component):
                return node.ayns.remove_child(component)

            return ComposedNode.ayns._remove_node(self, remove_fn, *path)

        def replace_node(self, new_node, *path):
            raise NotImplementedError()

        def filter_nodes(self, condition, prefix=None):
            prefix = ComposedNode.get_list_path(prefix, check_types=False) or NodePath()
            to_del = []
            to_re_set = []
            for name, child in self.ayns.named_children():
                child_path = prefix + [name]
                keep = False
                if condition(child_path, child):
                    keep = True
                if isinstance(child, ComposedNode): #not child.ayns.is_leaf:
                    possibly_new_child = child.ayns.filter_nodes(condition, prefix=child_path)
                    keep = keep and bool(possibly_new_child)
                    if keep and possibly_new_child is not child:
                        to_re_set.append(name, possibly_new_child)

                if not keep:
                    to_del.append(name)

            for name in reversed(to_del):
                self.ayns.remove_child(name)
            for name, child in to_re_set:
                self.ayns.set_child(name, child)

            return self

        def map_nodes(self, map_fn, prefix=None, cache_results=True, cache=None, leafs_only=True):
            prefix = ComposedNode.get_list_path(prefix, check_types=False) or NodePath()
            to_re_set = []
            if cache_results and cache is None:
                cache = {}

            for name, child in self.ayns.named_children():
                if cache_results and id(child) in cache:
                    possibly_new_child = cache[id(child)]
                else:
                    child_path = prefix + [name]
                    possibly_new_child = child
                    if isinstance(child, ComposedNode): #not child.ayns.is_leaf:
                        possibly_new_child = possibly_new_child.ayns.map_nodes(map_fn, prefix=child_path, cache_results=cache_results, cache=cache, leafs_only=leafs_only)
                        #if not leafs_only:
                        #    possibly_new_child = map_fn(child_path, possibly_new_child)
                    else:
                        possibly_new_child = map_fn(child_path, possibly_new_child)
                        
                if possibly_new_child is not child:
                    to_re_set.append((name, possibly_new_child))

                if cache_results:
                    from ..utils import persistent_id
                    cache[persistent_id(child)] = possibly_new_child

            for name, child in to_re_set:
                self.ayns.set_child(name, child)

            ret = self
            if not leafs_only:
                ret = map_fn(prefix, self)

            return ret

        def nodes_with_paths(self, prefix=None, recursive=True, include_self=False, allow_duplicates=True):
            memo = set()
            prefix = ComposedNode.get_list_path(prefix, check_types=False) or NodePath()
            if include_self:
                memo.add(id(self))
                yield prefix, self

            for name, child in self._children.items():
                if child is None or (id(child) in memo and not allow_duplicates):
                    continue
                child_path = prefix + [name]
                if not recursive or not isinstance(child, ComposedNode):
                    memo.add(id(child))
                    yield child_path, child
                else:
                    for path, node in child.ayns.nodes_with_paths(prefix=child_path, recursive=recursive, include_self=True):
                        if id(node) in memo and not allow_duplicates:
                            continue
                        memo.add(id(node))
                        yield path, node

        def nodes(self, recursive=True, include_self=False, allow_duplicates=False):
            for _, node in self.ayns.nodes_with_paths(recursive=recursive, include_self=include_self, allow_duplicates=allow_duplicates):
                yield node

        def nodes_paths(self, prefix=None, recursive=True, include_self=False):
            for path, _ in self.ayns.nodes_with_paths(recursive=recursive, include_self=include_self, allow_duplicates=True):
                yield path


        def named_children(self, allow_duplicates=True):
            memo = set()
            for name, child in self._children.items():
                if not allow_duplicates:
                    if id(child) in memo:
                        continue
                    memo.add(id(child))
                yield name, child

        def children(self, allow_duplicates=True):
            for _, child in self.ayns.named_children(allow_duplicates=allow_duplicates):
                yield child

        def children_names(self):
            for key in self._children.keys():
                yield key

        def children_count(self, allow_duplicates=True):
            if allow_duplicates:
                return len(self._children)
            else:
                return len(set(self._children.values()))

        def clear(self):
            self._children.clear()

        def premerge(self, into=None):
            ''' Called when `self` (top-level) is about to be merged into `into`
            '''
            return self.ayns.map_nodes(lambda path, node: node.ayns.on_premerge(path, into), cache_results=True, leafs_only=False)

        def preprocess(self, builder):
            return self.ayns.map_nodes(lambda path, node: node.ayns.on_preprocess(path, builder), cache_results=True, leafs_only=False)

        def evaluate(self):
            return self.ayns.map_nodes(ConfigNode.ayns.evaluate_node, cache_results=True, leafs_only=False)

        def merge(self, other):
            other.ayns.premerge(self)
            if other.ayns.delete:
                def maybe_keep(path, node):
                    other_node = other.ayns.get_first_not_missing_node(path)
                    return node.ayns.has_priority_over(other_node)

                self.ayns.filter_nodes(maybe_keep)
                if not self._children and other.ayns.has_priority_over(self, if_equal=True):
                    return other

            for key, value in other._children.items():
                child = self.ayns.get_child(key, None)
                if child is None:
                    self.ayns.set_child(key, value)
                else:
                    merge = False
                    #if not child.ayns.is_leaf: #and type(child) == type(value):
                    try:
                        possibly_new_child = child.ayns.merge(value)
                        merge = True
                    except (TypeError, AttributeError):
                        pass

                    if merge:
                        if not possibly_new_child and not possibly_new_child.ayns.has_priority_over(value):
                            self.ayns.remove_child(key)
                        elif possibly_new_child is not child:
                            self.ayns.set_child(key, possibly_new_child)
                            possibly_new_child._metadata = { **child._metadata, **possibly_new_child._metadata }
                    else:
                        if value.ayns.has_priority_over(child, if_equal=True):
                            if not value and value.ayns.explicit_delete:
                                self.ayns.remove_child(key)
                            else:
                                self.ayns.set_child(key, value)
                                value._metadata = { **child._metadata, **value._metadata }

            if other.ayns.has_priority_over(self, if_equal=True):
                self._merge_mode = other._merge_mode
                self._delete = other._delete
                self._implicit_delete = other._implicit_delete
                self._metadata = { **self._metadata, **other._metadata }
            else:
                self._metadata = { **other._metadata, **self._metadata }

            return self

        @staticproperty
        @staticmethod
        def is_leaf():
            return False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_children']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    @staticmethod
    def _recreate(cls):
        new = cls.__new__(cls)
        new._children = {}
        return new

    def __reduce__(self):
        state = self.__getstate__()
        lit = None
        dit = None
        if isinstance(self, list):
            lit = iter(self)
        elif isinstance(self, dict):
            dit = iter(self.items())
        return ComposedNode._recreate, (type(self), ), state, lit, dit



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections.abc as cabc

from .composed import ComposedNode
from ..namespace import namespace
from ..utils import Bunch


class ConfigDict(ComposedNode, dict):
    def __init__(self, value=None, **kwargs):
        value = value if value is not None else {}
        #ComposedNode.maybe_inherit_flags(value, kwargs)
        ComposedNode.__init__(self, children=value, **kwargs)
        dict.__init__(self, self._children)

    def _set(self, name, value):
        if name in dir(type(self)):
            raise ValueError(f'Cannot add a child node with name {name!r} as it would shadow a class method/attribute: {getattr(type(self), name)}')
        value = ComposedNode.ayns.set_child(self, name, value)
        dict.__setitem__(self, name, value)
        return value

    def _del(self, name):
        ret = ComposedNode.ayns.remove_child(self, name)
        dict.__delitem__(self, name)
        return ret

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return ComposedNode.__setattr__(self, name, value)

        return self._set(name, value)

    def __getattr__(self, name):
        if not ComposedNode.ayns.has_child(self, name):
            raise AttributeError(f'Object {type(self).__name__!r} does not have attribute {name!r}')

        return ComposedNode.ayns.get_child(self, name)

    def __delattr__(self, name):
        if name.startswith('_'):
            return ComposedNode.__delattr__(self, name)

        self._del(name)

    def __setitem__(self, name, value):
        if isinstance(name, str) and name.startswith('_'):
            return dict.__setitem__(self, name, value)

        return self._set(name, value)

    def __delitem__(self, name):
        if isinstance(name, str) and name.startswith('_'):
            return dict.__delitem__(self, name)

        self._del(name)

    def __contains__(self, name):
        return self.ayns.has_child(name) # pylint: disable=no-value-for-parameter

    @namespace('ayns')
    def set_child(self, name, value):
        self._set(name, value)

    @namespace('ayns')
    def remove_child(self, name):
        return self._del(name)

    def clear(self):
        ComposedNode.ayns.clear(self)
        dict.clear(self)

    def setdefault(self, key, value):
        if key not in self:
            return self._set(key, value)
        return self[key]

    def pop(self, k, *d):
        val = dict.pop(self, k, *d)
        if self.ayns.has_child(k): # pylint: disable=no-value-for-parameter
            ComposedNode.ayns.remove_child(self, k)
        return val

    def popitem(self, k, d=None):
        val = dict.popitem(self, k, d=d)
        if self.has_child(k):
            ComposedNode.ayns.remove_child(self, k)
            #val.set_parent(None, None)
        return val

    @namespace('ayns')
    def merge(self, other):
        if not isinstance(other, dict) and not isinstance(other, cabc.Sequence):
            raise TypeError('Dict or sequence expected')

        return super().ayns.merge(other)

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        return Bunch((key, ctx.evaluate_node(value, path+[key])) for key, value in self.ayns.named_children())

    def __repr__(self, simple=False):
        dict_repr = '{' + ', '.join([f'{n!r}: {c.__repr__(simple=True)}' for n, c in self.ayns.named_children()]) + '}' # pylint: disable=no-value-for-parameter
        if simple:
            return type(self).__name__ + dict_repr

        node = ComposedNode.__repr__(self)
        return node + dict_repr

    def __str__(self):
        def strify(kv):
            k, v = kv
            return repr(k) + ': ' + v.__repr__(simple=True)

        return type(self).__name__ + '{' + ', '.join(map(strify, self.items())) + '}'

    def _get_value(self):
        return self

    def _get_native_value(self):
        return dict((k,v.ayns.native_value) for k,v in self.items())

    def _set_value(self, other):
        self.clear()
        for name, child in other.items():
            self.ayns.set_child(name, child) # pylint: disable=no-value-for-parameter



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .eval import EvalNode


class FStrNode(EvalNode):
    ''' Implements ``!fstr`` tag.

        The f-string node can be used to construct dynamic strings
        in an convenient way - exactly the same as in standard Python.
        This behaviour is achieved by reducing the f-string node
        to a special case of an eval node.
        More specifically, the string associated with the f-string node
        is is transformed to form a valid Python f-string which is later
        evaluated with ``eval`` by the means of the :py:class:`EvalNode` class.
        This can be summarised as::

            def FStrNode(fmt):
                return EvalNode("f'{}'".format(escape(fmt)))

        where ``fmt`` should be the content of an f-string, without extra quotes
        and the leading ``f``. 

        Supported syntax::

            !fstr fmt       # explicit form
            f'fmt'          # implicit form, eq. to: !fstr fmt
            f"fmt"          # implicit form, eq. to: !fstr fmt

        Merge behaviour:

            The same as standard string node.
    '''
    def __init__(self, fstr, **kwargs):
        if len(fstr) < 3 or fstr[0] != 'f' or fstr[1] not in ['"', "'"] or fstr[1] != fstr[-1]:
            raise ValueError(f'Invalid f-string: {fstr!r}')
        super().__init__(fstr, persistent_namespace=False, **kwargs)



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .scalar import ConfigScalar
from ..namespace import namespace, staticproperty


class PrevNode(ConfigScalar(str)):
    def __init__(self, ref, **kwargs):
        super().__init__(ref)

    @namespace('ayns')
    def on_premerge(self, path, into):
        node = into.ayns.remove_node(self)
        if node is None:
            raise KeyError(f'Node {str(self)!r} does not exist in the previous context (possibly deleted?) - while processing a {type(self).__name__!r} node at {path!r}')
        return node

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def tag():
        return '!prev'



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dict import ConfigDict
from .function import FunctionNode
from ..namespace import namespace, staticproperty
from ..utils import import_name

from functools import partial


class BindNode(FunctionNode):
    ''' A class implementing ``!bind`` node.

        The bind tag creates a dict node extended with a string.
        The children of the bind node are interpreted as arguments
        to a callable entity identified by the string.

        In most cases, the bind node can be treated like a normal
        dict in which case the string part is "hidden" as it's not
        exposed through a standard dict API.
        However, due to the children nodes being connected to a particular
        value of the string, the merging behaviour of the bind node is
        slightly different. Specifically, if the entity's name changes
        due to merging, bind node discards any existing children.
        Please see the table below for details.

        The children can be identified by either string or int keys.
        In the first case, their values should match names of the function's
        arguments (analogical to standard keyword arguments). In the second
        case, the integers can be used to specify positions of arguments -
        they do not need to be continues as long as overall a valid
        bind (and consecutive call) is formed. If both a positional
        and keyword arguments match the same argument in the function's
        signature, an error is raised.

        When evaluated, the bind node returns a ``functools.partial``
        obtained by binding the arguments kept in its children to the
        entity identified by its string part (following semantics expplained
        briefly above). Specifically, the behaviour could be summarized as::

            f = import_name(bind_node.ayns.func)
            return functools.partial(f, **bind_node)

        Supported syntax::

            !bind:name { args }
            !bind:name [args]    # eq. to: !bind:name { enumerate(args) }
            !bind name           # eq. to: !bind:name {}


        Merge behaviour:

            ==================  ================================================================================================================================
            Case                Behaviour
            ==================  ================================================================================================================================
            ``dict <- Bind``    behaves as ``dict <- !del dict``, unless delete is explicitly set to False
            ``None <- Bind``    ``Bind``
            ``Bind1 <- Bind2``  ``Bind2`` if target entity changes, otherwise ``dict1 <- dict2``
            ``Bind <- dict``    Update ``Bind``'s arguments without changing target function
            ``Bind <- list``    Update ``Bind``'s arguments (analogical to ``dict <- list``) without changing target function
            ``Bind <- str``     if ``str`` is different than the current target function's name, update the name and remove all children, otherwise no effect
            ==================  ================================================================================================================================

    '''

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        _func = self._func
        if isinstance(_func, str):
            _func = import_name(_func)
        args = ConfigDict.ayns.on_evaluate(self, path, ctx)
        p, kw_p, kw = FunctionNode._resolve_args(_func, args)
        return partial(_func, *p, **kw_p, **kw)

    @namespace('ayns')
    @property
    def tag(self):
        _func = self._func
        if not isinstance(_func, str):
            _func = self._func.__module__ + '.' + self._func.__name__
        return '!bind:' + _func



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .scalar import ConfigScalar
from .composed import ComposedNode
from ..namespace import namespace, staticproperty


class XRefNode(ConfigScalar(str)):
    def __init__(self, value, **kwargs):
        super().__init__(value, **kwargs)

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        chain = [ComposedNode.get_str_path(path)]
        curr = self
        while isinstance(curr, XRefNode):
            try:
                ref = ctx.ayns.get_node(curr)
            except KeyError:
                msg = f'Referenced node {str(curr)!r} is missing, while following a chain of references: {chain}'
                raise ValueError(msg) from None

            chain.append(str(curr))
            curr = ref
        assert curr is not self
        return ctx.evaluate_node(curr, prefix=chain[-1])
        #return curr.ayns.on_evaluate(path, ctx)

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def tag():
        return '!xref'



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .node import ConfigNode
from ..namespace import namespace, staticproperty

import os
import collections.abc as cabc


class IncludeNode(ConfigNode):
    ''' Implements ``!include`` tag.

        The include node can be used to include content of another
        config file in any part of the currently parsed file.

        The content of the node (string) will be treated as a filename
        of the file which will be included in the place of the node.
        This filename can either be absolute or relative.
        In the later case, a specific file will be looked up in directories
        defined by a builder object which was used to build the config - please
        see :py:meth:`awesomeyaml.Builder.get_lookup_dirs`.
        The standard implementation looks up files with respect to the source file
        of the include node (preferred) or the current working directory.

        Include nodes can also be constructed with a sequence of names.
        In that case, the files are read and merged together in order to
        form the final config object which will be used to replace the include
        node. See examples below.

        Supported syntax::

            !include [files]    # long form
            !include file       # short form, eq. to: !include [file]

        Merge behaviour:

            Include nodes are somewhat special because their lifespan is very short.
            In fact, the standard behaviour is to process them even before merging
            is performed (since they define content which will be merged) during a stage
            called "preprocessing".
            Preprocessing is very similar to evaluating in a sense that it also allows
            a node to change itself. The main difference is the fact that during preprocessing
            a node is not given access to other nodes and therefore should be self-contained.
            Because of this short lifespan, include nodes don't have explicitly defined
            merging behaviour. However, should one try to merge a config tree with unprocessed
            include node, it would behave following the most common merging behaviour.

        Examples:
            The following snippets are equivalent.
            Python only::

                cfg = awesomeyaml.Config.build('file1', 'file2', 'file3')

            Aggregated in a single file::

                # in yaml (content of master_file)
                ---
                !include file1
                ---
                !include file2
                ---
                !include file3

                # in Python
                cfg = awesomeyaml.Config.build('master_file')

            The same using long form:

                # in yaml (content of master_file)
                ---
                !include [file1, file2, file3]

                # in Python
                cfg = awesomeyaml.Config.build('master_file')

    '''
    def __init__(self, filenames, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not isinstance(filenames, cabc.Sequence) or isinstance(filenames, str):
            filenames = [filenames]

        for filename in filenames:
            if not isinstance(filename, str):
                raise ValueError(f'Include node expects a string parameter with a filename to read, but got: {type(filename)}')

        self.filenames = [os.path.expanduser(f) for f in filenames]

    @namespace('ayns')
    def on_preprocess(self, path, builder):
        subbuilder = builder.get_subbuilder(path)
        missing = []
        for filename in self.filenames:
            found = False
            for lookup_dir in subbuilder.get_lookup_dirs(self._source_file):
                file = os.path.normpath(os.path.join(lookup_dir, filename))
                try:
                    subbuilder.add_source(file, raw_yaml=False)
                    found = True
                except FileNotFoundError:
                    continue

                if found:
                    break

            if not found:
                missing.append(filename)

        if missing:
            raise FileNotFoundError({ 'missing': missing, 'lookup_dirs': list(subbuilder.get_lookup_dirs(self._source_file)), 'source': self._source_file })

        return subbuilder.build().ayns.on_preprocess(path, builder)


    @namespace('ayns')
    @staticproperty
    @staticmethod
    def tag():
        return '!include'

    @namespace('ayns')
    @property
    def value(self):
        return {
            'filenames': self.filenames,
            'ref_file': self.ref_file
        }



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .list import ConfigList
from ..namespace import namespace, staticproperty


class StreamNode(ConfigList):
    def __init__(self, builder, **kwargs):
        kwargs.setdefault('delete', False)
        super().__init__(builder.stages, **kwargs)
        self._default_delete = False
        self.builder = builder

    @property
    def stages(self):
        return self.builder.stages

    @namespace('ayns')
    def on_premerge(self, path, into):
        self.clear()
        self.builder.flatten()
        self.append(self.builder.stages[0])
        return self.builder.stages[0].ayns.on_premerge(path, into)

    #@namespace('ayns')
    #@staticproperty
    #@staticmethod
    #def is_leaf():
    #    return True



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .list import ConfigList
from ..namespace import namespace, staticproperty

from collections.abc import Sequence


class AppendNode(ConfigList):
    ''' A class implementing ``!append`` node.

        The append tag creates a list node which appends its
        content to another list on merge, instead of overwriting it.

        Supported syntax::

            !append [list]

        Merge behaviour:

            ==================  =================================
            Case                Behaviour
            ==================  =================================
            ``A <- Append``     ``return A.extend(Append)``
            ``None <- Append``  ``raise KeyError``
            otherwise           behaves as :py:class:`ConfigList`
            ==================  =================================

    '''
    def __init__(self, value, **kwargs):
        if not isinstance(value, Sequence) or isinstance(value, str) or isinstance(value, bytes):
            value = [value]
        super().__init__(value, **kwargs)

    @namespace('ayns')
    def on_premerge(self, path, into):
        if into is None:
            return ConfigList(self)

        node = into.ayns.remove_node(path)
        if node is None:
            raise KeyError(f'Node {str(self)!r} does not exist in the previous context (possibly deleted?) - while processing a {type(self).__name__!r} node at {path!r}')
        node.extend(self)
        return node

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def tag():
        return '!append'



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dict import ConfigDict
from ..namespace import namespace, staticproperty
from ..utils import import_name


class FunctionNode(ConfigDict):
    def __init__(self, func, args=None, **kwargs):
        ''' Func can be either str naming a function, or a pair (name, args) in which case args should be None
        '''
        if not isinstance(func, str):
            if isinstance(func, tuple):
                if args is not None:
                    raise ValueError('"args" passed directly and together with "func" resulting in an ambiguous assignment')
                self._func, args = func
            else:
                self._func = func
        else:
            self._func = func

        if args is not None and not isinstance(args, dict):
            if isinstance(args, list) or isinstance(args, tuple):
                args = { idx: val for idx, val in enumerate(args) }
            else:
                args = { 0: args }

        if not kwargs.get('delete', None):
            kwargs.setdefault('delete', True)
        super().__init__(args, **kwargs)
        self._default_delete = True

    def __bool__(self):
        return bool(self._func)

    @namespace('ayns')
    def merge(self, other):
        if isinstance(other, str):
            self._func = other
            self.clear()
            return self

        if isinstance(other, FunctionNode) and type(self) != type(other):
            raise TypeError('Conflicting FunctionNode subclasses')

        try:
            self._func = other._func
        except AttributeError:
            pass

        return super().ayns.merge(other)

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        raise NotImplementedError()

    @namespace('ayns')
    def represent(self):
        return self.ayns.tag, self.ayns.get_node_info_to_save(), super()._get_value()

    @namespace('ayns')
    @property
    def func(self):
        return self._func

    def _get_value(self):
        return (self._func, super()._get_value())

    def _get_native_value(self):
        return (self._func, super()._get_native_value())

    def _set_value(self, value):
        self._func, super_val = value
        return super()._set_value(super_val)

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def is_leaf():
       return False

    @namespace('ayns')
    @property
    def tag(self):
        raise NotImplementedError()

    @staticmethod
    def _resolve_args(func, args):
        positional_args = { key: value for key, value in args.items() if isinstance(key, int) }
        if not positional_args:
            return [], {}, args

        keyword_args = { key: value for key, value in args.items() if isinstance(key, str) }
        assert len(positional_args) + len(keyword_args) == len(args)

        import inspect
        sig = inspect.signature(func)
        params = list(sig.parameters.values())
        idx_to_name = []
        for p in params:
            if p.kind == inspect.Parameter.VAR_POSITIONAL:
                break
            idx_to_name.append(p.name)

        idx = 0
        unpack = []
        while True:
            if idx not in positional_args:
                break
            unpack.append(positional_args.pop(idx))
            idx += 1

        kw_positional_args = {}
        for idx, value in positional_args.items():
            if idx >= len(idx_to_name):
                raise ValueError(f'Cannot resolve argument at position {idx} for function: {func} with signature {sig}')
            kw_positional_args[idx_to_name[idx]] = value

        return unpack, kw_positional_args, keyword_args



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys


class persistent_id(int):
    ''' The purpose of this class is to replace built-in ``id`` function
        in cases when keeping an id of an object should prolong the object's
        life, mostly to ensure that the returned id will be unique within a certain
        scope (to provide an easy way of identifying objects).
        This class is a simple extension of the `int` class which additionally
        to the normal value holds a reference to the object whose id it holds,
        therefore effectively preventing it from being deallocated and ensuring
        that the id will continue to uniquely identify that object.
        The downside of this approach is the fact that the objects will
        remain in memory for as long as its persistent_id is in use - therefore
        it should be used in cases when memory is not an issue and correctness of
        unique identification is preferred.
        
        An example use case could be having a cache (memo) of objects already
        processed, where the incoming objects are managed (created etc.) by a user,
        for example with a generator. The generator might but might not return the same
        object multiple times in which case we don't want to process it multiple times.
        Since we don't know in advance if an object will be given to us multiple times,
        we need to ensure that we can identify a situation when it happens as the objects arrive.
        A simple idea is to keep track of ids of objects which have already been processed in
        a cache/memo, using for example a set::

            def process_all(objs):
                memo = set()
                for obj in objs:
                    if id(obj) not in memo:
                        memo.add(id(obj))
                        yield process(obj)

        where ``process`` is potentially expensive processing of objects in ``objs``.
        However, if it is possible for an object to be deleted after processing (i.e. no one stores
        a reference to it) the id might be repeated easily but for a different object, creating
        a false impression that the object has already been processed, where in fact it was
        a different one which just happened to had the same id.
        The following code illustrates that::

            class MyObject():
                def __init__(self, value):
                    self.value = value

            def obj_gen(num_objs):
                for i in range(num_objs):
                    yield MyObject(i)

            def process(obj):
                # expensive computations
                return obj.value ** 2

            def process_all(objs):
                memo = set()
                for obj in objs:
                    if id(obj) not in memo:
                        memo.add(id(obj))
                        yield process(obj)

            processed = list(process_all(obj_gen(100)))
            assert len(processed) == 100

        When running the above code, it's possible that only a couple of first objects will be processed
        (in my case first 2) because later objects will be allocated at the place of previous ones which will
        trick the ``memo`` within ``process``. Therefore it's important to keep a reference to all objects
        whose ids are kept to ensure their uniqueness. A couple of different solutions could be used to
        address it, e.g.: instead of a set one could use dict and store pairs ``id(obj): obj`` within it;
        or simply precompute all objects ahead of execution and keep them in a separate container.
        Both approaches have very similar cost in terms of memory consumption  as the solution presented here
        (prevent all objects from being deallocated), but at the same time they have some additional drawbacks:

         - the approach with storing the original object in a set can add some hassle if a dict is already in use
           to store results of processing of the object, i.e.::

                def process_all(objs):
                    memo = {}
                    for obj in objs:
                        if id(obj) not in memo:
                            result = process(obj)
                            memo[id(obj)] = result
                        else:
                            result = memo[id(obj)]

                        yield result

           the solution would be to either store pairs ``id(obj): (obj, process(obj))`` or have two memos which in both cases pollutes the code;
         - the second solution requires precomputing all values ahead of time which will not work nicely when chaining generators.

        The proposed solution requires simply changing ``id`` to ``persistant_id`` which is clean and nice::

            class MyObject():
                def __init__(self, value):
                    self.value = value

            def obj_gen(num_objs):
                for i in range(num_objs):
                    yield MyObject(i)

            def process(obj):
                # expensive computations
                return obj.value ** 2

            def process_all(objs):
                memo = {}
                for obj in objs:
                    if id(obj) not in memo:
                        result = process(obj)
                        memo[persistent_id(obj)] = result
                    else:
                        result = memo[id(obj)]

                    yield result

            processed = list(process_all(obj_gen(100)))
            assert len(processed) == 100

    '''

    def __new__(cls, obj):
        return int.__new__(cls, id(obj))

    def __init__(self, obj):
        ''' Return id of the ``obj`` which holds a reference to it, ensuring that no other
            object will use the same id as long as the ``persistent_id`` object exists.
        '''
        self._ref = obj


def pad_with_none(*args, minlen=None):
    if minlen is None or len(args) >= minlen:
        return args
    return (*args, *([None] * (minlen - len(args))))


def import_name(symbol_name):
    if not symbol_name or symbol_name.endswith('.'):
        raise ValueError(f'Invalid target name: {symbol_name}')

    elements = symbol_name.split('.')
    current = None
    try_import = True
    exceptions = []

    def _build_import_exception(symbol, last, excs):
        flat_excs = []
        for e in excs:
            this_ex_chain = []
            while e is not None:
                this_ex_chain.insert(0, e)
                e = getattr(e, '__cause__', None)

            flat_excs.extend(this_ex_chain)

        ex = None
        for e in flat_excs:
            if ex is not None:
                e.__cause__ = ex
            ex = e

        if ex is not None:
            import traceback
            ex = ''.join(traceback.format_exception(type(ex), ex, ex.__traceback__))
        else:
            ex = '<No information>'

        return ImportError(f'Cannot find an entity named: {symbol!r}, last found element was: {last}, see exception(s) below for potential reasons what could have gone wrong:\n\n{ex}')

    for element in elements:
        if try_import:
            import importlib
            if current:
                try:
                    current = importlib.import_module('.' + element, package=current.__name__)
                    continue
                except ImportError as e:
                    exceptions.append(e)
                    try_import = False
            else:
                try:
                    current = importlib.import_module(element)
                    continue
                except ImportError as e:
                    exceptions.append(e)
                    try_import = False

        if current is not None:
            try:
                current = getattr(current, element)
                continue
            except AttributeError as e:
                exceptions.append(e)
                pass

        if current is None and len(elements) == 1:
            import builtins
            try:
                current = getattr(builtins, element)
                continue
            except AttributeError as e:
                exceptions.append(e)
                pass

        raise _build_import_exception(symbol_name, current, exceptions)

    return current


class Bunch(dict):
    def __init__(self, other):
        super().__init__(other)

    def __getattr__(self, name):
        if name not in self:
            raise AttributeError(f'Object {type(self).__name__!r} does not have attribute {name!r}')
        return self[name]

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return super().__setattr__(name, value)

        if name in self.__dict__:
            raise ValueError('Name conflict!')

        self[name] = value

    def __delattr__(self, name):
        try:
            super().__delattr__(name)
        except AttributeError:
            del self[name]


class LazyModule():
    def __init__(self, module):
        self.module = module

    def __getattr__(self, name):
        return getattr(self.module, name)


def add_module_properties(module_name, properties):
    module = sys.modules[module_name]
    replace = False
    if isinstance(module, LazyModule):
        lazy_type = type(module)
    else:
        lazy_type = type('LazyModule({})'.format(module_name), (LazyModule,), {})
        replace = True

    for name, prop in properties.items():
        setattr(lazy_type, name, prop)

    if replace:
        sys.modules[module_name] = lazy_type(module)


def python_is_at_least(major, minor):
    return sys.version_info[0] > major or (sys.version_info[0] == major and sys.version_info[1] >= minor)


def python_is_exactly(major, minor):
    return sys.version_info[0] == major and sys.version_info[1] == minor



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import contextlib
import collections.abc as cabc

from .nodes.node import ConfigNode


class Builder():
    ''' A builder class responsible for merging multiple yaml configs into a single
        node.

        When populating a builder object, it will maintain a list of 'stages' with each
        stage being a single yaml document. :py:meth:`add_source` and :py:meth:`add_multiple_sources` can be
        used to add new stages - the first function should be given a source of yaml documents
        (e.g. a file) and will greedily read its content appending new documents to the list
        of stages (in order of reading). For more information about what can be a source
        for yaml documents, see :py:meth:`add_source`. :py:meth:`add_multiple_sources` simply proides a convenient
        way of adding multiple sources without calling :py:meth:`add_source` multiple times - its delegates
        all actual work to :py:meth:`add_source`.
        
        After all relevant stages have been added, they can be later merged into a single
        node by calling :py:meth:`build`.
    '''
    def __init__(self):
        ''' Creates an empty builder. Yaml documents can then be added with calls to :py:meth:`add_source`
            and :py:meth:`add_multiple_sources`.
        '''
        self.stages = []
        self._current_file = None
        self._current_stage = None

    @contextlib.contextmanager
    def current_stage(self, i):
        ''' A context manager used to mark a specific stage as 'currently being preprocessed'.

            The marking doesn't have any specific meaning imposed by this function, its responsibility
            of its user to interpret it.
            The function is intended to be used internally but is exposed here in case someone would
            like to derive from this class to customize preprocessing/merging of yaml documents.

            At the moment of writing this, the marking of the active stage is only used when recursively
            constructing subbuilders - which is used when recursively preprocessing include nodes (when a subbuilder
            is created, it reads the current stage id and saves it in order to be able to reconstruct
            a chain of includes, e.g. to report errors).

            Arguments:
                i (int, None) : index of the stage to be marked as active (can be ``None`` to indicate that no stage
                    is currently being preprocessed).

            Returns:
                A context manager which sets the current stage to ``i`` when entered and restored the previous
                values on exit.
        '''
        old = self._current_stage
        self._current_stage = i
        yield
        self._current_stage = old

    def get_next_stage_idx(self):
        ''' Returns index of an 'about-to-be-added' stage. This is basically always equal to the current length of the list of stages.
            The function is provided to enable easy change of this logic in case it's needed in the future.
        '''
        return len(self.stages)

    def get_current_stage_idx(self):
        ''' Return index of the stage 'currently being preprocessed', or ``None`` if no stage is being preprocessed.
            See :py:meth:`current_stage` and :py:meth:`preprocess` for more information about what 'being preprocessed' means.
        '''
        return self._current_stage

    def get_current_file(self):
        ''' Get name of the file currently being read - used by the yaml parser to include informations
            about origins of each node (especially important for include nodes).
            This can be ``None`` if the yaml content currently being parsed does not originate from a file.
        '''
        return self._current_file

    def add_multiple_sources(self, *sources, raw_yaml=None, filename=None):
        ''' Adds multiple sources using :py:meth:`add_source` function. ``raw_yaml`` and ``filename`` have the same meaning
            as in :py:meth:`add_source` and can be either a single scalar, in which case its value is broadcasted
            and applied to all elements in ``*sources``, or they can be sequences of equal length if different
            values should be applied to different elements in ``*sources``. It's perfectly fine for one of them
            to be a scalar value and the second one to be a sequence.

            Returns:
                ``None``
        '''
        def sanitize(arg, arg_name):
            if not isinstance(arg, cabc.Sequence) or isinstance(arg, str) or isinstance(arg, bytes):
                arg = [arg] * len(sources)
            else:
                #if isinstance(arg, str) or isinstance(arg, bytes):
                #    raise ValueError(f"{arg_name!r}: Expected sequence but not str or bytes")
                if len(arg) != len(sources):
                    raise ValueError(f"Length of 'sources' and {arg_name!r} must match")

            return arg

        raw_yaml = sanitize(raw_yaml, 'raw_yaml')
        filename = sanitize(filename, 'filename')
        for source, raw, fname in zip(sources, raw_yaml, filename):
            self.add_source(source, raw_yaml=raw, filename=fname)

    def add_source(self, source, raw_yaml=None, filename=None):
        ''' Parse a stream of yaml documents specified by ``source`` and add them to the list of stages.
            The documents stream can be provided either directly or read from a file - the behaviour is determined
            by ``source`` and ``raw_yaml`` arguments.

            Args:
                source : either a string or a file object. If it's a file object, it will be passed directly to the :py:func:`awesomeyaml.yaml.parse`,
                           otherwise it will be treated either as a filename or yaml to be parsed, according to ``raw_yaml``.
                raw_yaml : controls how ``source`` is interpreted if it is passed as string, possible cases are:

                            - if ``raw_yaml`` is set to ``None`` specifically (default), the function will try guessing whether ``source`` is a
                              name of a file or a yaml string to be parsed directly. In order to do that, it will behave as if ``raw_yaml`` was set to ``False``
                              and in case ``FileNotFoundError`` is raised it will fallback to the case when ``raw_yaml`` is set to ``True``.
                            - if ``bool(raw_yaml)`` evaluates to ``False`` and is not ``None``, the function will attempt to open and read content of a file named ``source``,
                              raising an error if such a file does not exist
                            - if ``bool(raw_yaml)`` evaluates to ``True``, ``source`` is treated as a yaml string and passed directly to the :py:func:`awesomeyaml.yaml.parse`

            Returns:
                ``None``
        '''
        if raw_yaml and not isinstance(source, str):
            raise ValueError('source is expected to be string and contain yaml to be parsed when raw_yaml is set to True')

        if isinstance(source, pathlib.Path):
            source = str(source)

        if isinstance(source, str) and not raw_yaml:
            try:
                with open(os.path.expanduser(source), 'r') as f:
                    self._current_file = source
                    source = f.read()
            except (FileNotFoundError, OSError) as e:
                #OSError(22) is "Invalid argument"
                if type(e) is OSError and e.errno != 22:
                    raise
                if raw_yaml is not None:
                    raise

        try:
            if filename is not None:
                self._current_file = filename

            with ConfigNode.default_filename(self._current_file):
                from . import yaml
                for node in yaml.parse(source, self):
                    if node is not None:
                        self.stages.append(node)
        finally:
            self._current_file = None

    def build(self):
        ''' Preprocesses all stages and merges them to construct a single config node.
            This node is suitable to be passed to :py:class:`awesomeyaml.eval_context.EvalContext`
            in order to be evaluated.

            Returns:
                A py:class:`awesomeyaml.nodes.ConfigDict` node representing merged config.
        '''
        if not self.stages:
            return None

        self.preprocess()
        self.flatten()
        return self.stages[0]

    def preprocess(self):
        ''' Preprocesses all stages.

            'Preprocessing' is a building step when each stage has an opportunity to modify itself
            without involving other stages. This is done mostly to handle things (nodes) which might
            require some knowledge about other fields in the stage and therefore must be run
            after parsing is done, but should also be run before merging as they might change the merging
            outcome.

            Internally, this is achieved by triggering ``on_preprocess`` on all nodes within the stage.
            The bahaviour of ``on_preprocess`` will obviously vary from node to node, but as a representative
            example :py:class:`awesomeyaml.nodes.include.IncludeNode` can be used. An include node is used to trigger
            inclusion of another yaml file into the stage (which might also have include nodes, in which
            case they will be preprocessed recursively). Whereas they don't necessarily require other
            nodes to be preprocessed, there are a couple of reasons they are not resolved within the yaml
            parser:

                - first of all for debugability - having an explicit intermediate step when include nodes are
                  present but not yet preprocessed enables the user to see if the config building processes
                  goes as desired,
                - second of all, which is a little bit related to the first one, we don't want to make
                  yaml parser fail in case a user made an error and the included file cannot be found - that
                  is because parsing might be triggered before include nodes should be preprocessed (e.g. parsed
                  config can be send over to a different machine, when it's put together and includes are resolved).
                  Although currently no use-case requires this separation, it makes the overall design more flexible
                  and helps define clear boundaries between different parts of the package.

            Obviously, include nodes have to be preprocessed before merging happens as the result will be subject
            to merging, which makes them a nice candidate to be handled in during the preprocessing stage.
        '''
        i = 0
        while i < len(self.stages):
            _i = i
            with self.current_stage(i):
                stage = self.stages[i]
                new_stage = stage.ayns.preprocess(self)

                if new_stage is not stage:
                    try:
                        self.stages[i:i+1] = new_stage.stages
                        i += len(new_stage.stages)
                    except AttributeError:
                        self.stages[i] = new_stage
                        i += 1
                else:
                    i += 1

            assert _i != i, 'infinite loop?'

    def flatten(self):
        ''' Flattens the list of stages be iteratively merging all stages into a single one.
            Merging happens in the same order in which stages are stored, i.e.::

                stages[0].merge(stages[1]).merge(stages[2])...

        '''
        for stage in self.stages:
            if not isinstance(stage, dict):
                raise ValueError('Not all stages are dictionaries')

        new_stage = self.stages[0].ayns.premerge(None)
        if new_stage is not self.stages[0]:
            try:
                self.stages[0:1] = new_stage.stages
            except AttributeError:
                self.stages[0] = new_stage

        if len(self.stages) < 2:
            return

        for i in range(1, len(self.stages)):
            self.stages[0].ayns.merge(self.stages[i])

        self.stages = [self.stages[0]]

    def get_lookup_dirs(self, ref_point):
        ''' Yields of list of directories where files should be searched for.
            Used by include nodes.

            Arguments:
                ref_point : an optional reference file w.r.t. which the searching happens,
                    if provided, it's directory is always the first one to be returned by
                    the generator

            Yields:
                A list of lookup directories.
        '''
        if ref_point is not None:
            yield os.path.dirname(ref_point)
        yield os.getcwd()

    def get_subbuilder(self, requester):
        ''' Returns a subbuilder which can be used to recursively build substreams of yaml documents.

            A subbuilder is created to recursively trigger :py:meth:`preprocess`, for example when processing
            an include node.

            Arguments:
                requester : a path of the node which requests the subbuilder (mostly used for debugging)

            Returns:
                A :py:class:`awesomeyaml.builder.SubBuilder` object who parent is this builder.
        '''
        if self._current_stage is None:
            raise RuntimeError('SubBuilder requested when not processing any stage!')

        return SubBuilder(requester, self)


class SubBuilder(Builder):
    ''' A subbuilder which might be created to enable recursive building.
        It should be created with a call to :py:meth:`Builder.get_subbuilder`.

        A subbuilder managers its own list of stages which can be preprocessed independently
        from the stages of its parent (i.e. as a part of the preprocessing step of one of the
        parent's stages).

        Calling :py:meth:`build` on a subbuilder will trigger preprocessing but not flattening as this
        should be handled later when the parent enter its flattening step. Instead, a single
        :py:class:`awesomeyaml.nodes.stream.StreamNode` is created representing a list of stages which
        should be flattened later. When flattening, the stream node will be replaced with the result
        of flattening its list of stages.
    '''
    def __init__(self, srcnode, parent):
        ''' Creates a new subbuilder.

            Arguments:
                srcnode : a path to the node requesting the subbuilder (the node exists in parent)
                parent : a parent Builder
        '''
        super().__init__()
        self.requester = srcnode
        self.parent = parent
        self.stage = parent.get_current_stage_idx()

    def build(self):
        ''' Triggers :py:meth:`preprocess` on the list of stages handled by this subbuilder.
            Intended for recursive preprocessing.

            Returns:
                A :py:class:`awesomeyaml.nodes.stream.StreamNode` holding a list of preprocessed stages.
                The stages attached to it will be flatten with the parent stage of the stream
                node.
        '''
        from .nodes.stream import StreamNode
        self.preprocess()
        return StreamNode(self)

    def get_lookup_dirs(self, ref_point):
        return self.parent.get_lookup_dirs(ref_point)



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .composed import ComposedNode
from ..namespace import namespace, staticproperty
from .. import utils


class ConfigList(ComposedNode, list):
    def __init__(self, value=None, **kwargs):
        value = value if value is not None else []
        #ComposedNode.maybe_inherit_flags(value, kwargs)
        kwargs.setdefault('delete', True)
        ComposedNode.__init__(self, children={ i: v for i, v in enumerate(value) }, **kwargs)
        list.__init__(self, self._children.values())
        self._default_delete = True

    def _validate_index(self, index, allow_append=False):
        if not isinstance(index, int):
            raise TypeError(f'Index should be integer, got: {type(index)}')
        if abs(index) >= len(self) and (not allow_append or index != len(self)):
            raise IndexError('List index out of range')
        if index < 0:
            index = len(self) + index

        return index

    def _set(self, index, value, allow_append=False):
        index = self._validate_index(index, allow_append=allow_append)
        try:
            value = ComposedNode.ayns.set_child(self, index, value)
            if allow_append and index == len(self):
                list.append(self, value)
            else:
                list.__setitem__(self, index, value)
        except:
            ComposedNode.ayns.remove_child(self, index)
            raise

    def _del(self, index):
        index = self._validate_index(index)
        for i in range(index+1, len(self)):
            self[i-1] = self[i]

        ret = ComposedNode.ayns.remove_child(self, len(self) - 1)
        list.__delitem__(self, -1)
        return ret

    def _get(self, index, default=None, raise_ex=True):
        try:
            index = self._validate_index(index)
        except (IndexError, TypeError):
            if raise_ex:
                raise
            return default

        return list.__getitem__(self, index)

    def __setitem__(self, index, value):
        return self._set(index, value)

    def __delitem__(self, index):
        self._del(index)

    def __getitem__(self, index):
        return self._get(index, raise_ex=True)

    @namespace('ayns')
    def set_child(self, index, value):
        return self._set(index, value, allow_append=True)

    @namespace('ayns')
    def remove_child(self, index):
        return self._del(index)

    @namespace('ayns')
    def get_child(self, index, default=None):
        return self._get(index, default=default, raise_ex=False)

    def __contains__(self, value):
        return list.__contains__(self, value)

    def append(self, value):
        value = ComposedNode.ayns.set_child(self, len(self), value)
        list.append(self, value)

    def remove(self, value):
        self._del(self.index(value))

    def clear(self):
        ComposedNode.ayns.clear(self)
        list.clear(self)

    def extend(self, other):
        for val in other:
            self.append(val)

    if not utils.python_is_at_least(3, 7):
        # for python < 3.7 (i.e., 3.6 and older)
        # list subclasses are unpickled without calling
        # custom 'append' and/or 'extend' and rather
        # using "the fastest" possible option which is to
        # populate the built-in function type alone
        # (see: https://stackoverflow.com/questions/52333864/pickle-breaking-change-in-python-3-7)
        # Because of that we need to populate _children
        # of the composed node ourselves after the list has
        # been unpickled (which happens before __setstate__
        # is called).
        def __setstate__(self, state):
            super().__setstate__(state)
            assert not self._children
            self._children = { idx: child for idx, child in enumerate(self) }

    @namespace('ayns')
    def merge(self, other):
        if not isinstance(other, ComposedNode):
            raise TypeError('ComposedNode expected')

        if isinstance(other, dict):
            for key in other.ayns.children_names():
                self._validate_index(key)

        def keep_if_exists(path, node):
            if not node.ayns.delete:
                return True
            current = self.ayns.get_first_not_missing_node(path)
            return node.ayns.has_priority_over(current, if_equal=True)

        other.ayns.filter_nodes(keep_if_exists)

        return super().ayns.merge(other)

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        return list(ctx.evaluate_node(value, path+[key]) for key, value in self.ayns.named_children())

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def is_leaf():
        return False

    def __repr__(self, simple=False):
        list_repr = '[' + ', '.join([c.__repr__(simple=True) for c in self.ayns.children()]) + ']' # pylint: disable=no-value-for-parameter
        if simple:
            return type(self).__name__ + list_repr

        node = ComposedNode.__repr__(self)
        return node + list_repr

    def __str__(self):
        return type(self).__name__ + '[' + ', '.join(map(str, self)) + ']'

    def _get_value(self):
        return self

    def _get_native_value(self):
        return list(n.ayns.native_value for n in self)

    def _set_value(self, other):
        self.clear()
        for index, child in enumerate(other):
            self.set_child(index, child)



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .nodes.node import ConfigNode
from .nodes.composed import ComposedNode, NodePath
from .namespace import Namespace, NamespaceableMeta
from .utils import Bunch

import copy


class EvalContext(metaclass=NamespaceableMeta):
    ''' A class used to provide evaluation context for a top-level `awesomeyaml.nodes.ConfigDict`.
        After the top level node is put within a context, it can be evaluated to create `awesomeyaml.utils.Bunch`
        which will hold the evaluation result - evaluated nodes are no longer `awesomeyaml` nodes and can represent anything.
        The primary use-case for this class and the resulting `Bunch` object is to construct a `awesomeyaml.Config` object
        which basically combines the returned `Bunch` object with the original `ConfigDict`.
    '''

    class PartialChild(Bunch):
        def __init__(self, path, eval_ctx):
            self._path = path
            self._eval_ctx = eval_ctx
            super().__init__({})

        def __getitem__(self, key):
            if key not in self:
                return self._eval_ctx.cfg.ayns.get_node(self._path + [key])

            return super().__getitem__(key)

        def __getattr__(self, name):
            #if name not in self:
            #    raise AttributeError(f'Object {type(self).__name__!r} does not have attribute {name!r}')
            return self[name]

        def __setattr__(self, name, value):
            if name.startswith('_'):
                return super().__setattr__(name, value)

            if name in self.__dict__:
                raise ValueError('Name conflict!')

            self[name] = value

        def __delattr__(self, name):
            try:
                super().__delattr__(name)
            except AttributeError:
                del self[name]

    _default_eval_symbols = {}

    def __init__(self, config_dict, eval_symbols=None):
        ''' Arguments:
                config_dict : a `awesomeyaml.nodes.ConfigDict` object to be evaluated,
                    to construct a single `ConfigDict` from multiple sources
                    `awesomeyaml.builder.Builder` can be used.
                eval_symbols : a dict containing symbols which can be used when evaluating
                    ``config_dict``. The values from this argument will be used to update
                    the defaults from :py:meth:`get_default_eval_symbols`.
        '''
        self.cfg = config_dict
        self._removed_nodes = {}
        self._eval_cache = {}
        self._eval_paths = set()
        self._done = False
        self._eval_symbols = copy.copy(EvalContext._default_eval_symbols)
        if eval_symbols:
            self._eval_symbols.update(eval_symbols)

    class ayns(Namespace):
        def get_node(self, *path, **kwargs):
            def access_fn(node, component):
                return node[component]
            def query_fn(node, component):
                try:
                    _ = node[component]
                    return True
                except:
                    return False

            incomplete = kwargs.get('incomplete', False)
            kwargs['incomplete'] = False
            try:
                return ComposedNode.ayns._get_node(self.ecfg, access_fn, query_fn, *path, **kwargs)
            except KeyError:
                kwargs['incomplete'] = incomplete
                return self.cfg.ayns.get_node(*path, **kwargs)

        def remove_node(self, *path):
            def remove_fn(node, component):
                if isinstance(node, ComposedNode):
                    return node.ayns.remove_child(component)
                else:
                    node = node[component]
                    del node[component]
                    return node

            return ComposedNode.ayns._remove_node(self, remove_fn, *path)

        def named_children(self, allow_duplicates=True):
            memo = set()
            for name, child in self.ecfg.items():
                if not allow_duplicates:
                    if id(child) in memo:
                        continue
                    memo.add(id(child))
                yield name, child
            for name, child in self.cfg.ayns.named_children():
                if name not in self.ecfg:
                    yield name, child

    def get_evaluated_node(self, nodepath):
        nodepath = ComposedNode.get_list_path(nodepath, check_types=False) or NodePath()
        node = self.ayns.get_node(nodepath)
        if str(nodepath) in self._eval_paths:
            return node

        return self.evaluate_node(node, prefix=nodepath)

    def evaluate_node(self, cfgobj, prefix=None):
        if not isinstance(cfgobj, ConfigNode):
            return cfgobj

        prefix = ComposedNode.get_list_path(prefix, check_types=False) or NodePath()
        if str(prefix) in self._eval_paths:
            assert id(cfgobj) in self._eval_cache
            return self._eval_cache[id(cfgobj)]

        def maybe_evaluate(node):
            if id(node) in self._eval_cache:
                return self._eval_cache[id(node)]
            e = node.ayns.evaluate_node(prefix, self)
            self._eval_cache[id(node)] = e
            return e

        evaluated_parent = None
        if prefix:
            evaluated_parent = self.ayns.get_node(prefix[:-1])
            if isinstance(evaluated_parent, ConfigNode):
                evaluated_parent = self.evaluate_node(evaluated_parent, prefix=prefix[:-1])

            evaluated_parent[prefix[-1]] = EvalContext.PartialChild(prefix, self)

        evaluated_cfgobj = maybe_evaluate(cfgobj)
        if evaluated_parent is not None:
            evaluated_parent[prefix[-1]] = evaluated_cfgobj

        self._eval_paths.add(str(prefix))
        return evaluated_cfgobj

    def evaluate(self):
        ''' Returns:
                `awesomeyaml.utils.Bunch` representing evaluated config node.
        '''
        self._eval_cache.clear()
        self._eval_paths.clear()
        self.ecfg = {}
        return self.evaluate_node(self.cfg)

    @staticmethod
    def set_default_eval_symbols(symbols):
        ''' Sets the default symbols available to use when evaluating nodes.

            The symbols provided can be used by certain node types on evaluation.
            The defaults can be updated with values passed to the :py:meth:`__init__`
            method.

            Arguments:
                symbols : a lookup table (dict) of available symbols - for exampe, can be provided by
                    calling ``globals()`` in a module whose content one wishes to be available
                    during evaluation.

            Raises:
                TypeError: if ``symbols`` is not a dict.
        '''
        if not isinstance(symbols, dict):
            raise TypeError('dict expected')
        EvalContext._default_eval_symbols = symbols

    @staticmethod
    def get_default_eval_symbols():
        ''' Returns a dict with the default symbols available to use when evaluating nodes.

            The symbols provided can be used by certain node types on evaluation.
            The defaults can be updated with values passed to the :py:meth:`__init__`
            method.

            Returns:
                dict: symbols available by default
        '''
        return EvalContext._default_eval_symbols

    def get_eval_symbols(self):
        ''' Returns a dict with symbols available when evaluating nodes within this :py:class:`EvalContext`.

            Some node types might use this to evaluate their values.

            The returned value considers both the default symbols (see :py:meth:`set_default_eval_symbols`) and
            non-default ones passed to :py:meth:`__init__`.

            Returns:
                dict: symbols available in this evaluation context
        '''
        return self._eval_symbols



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .composed import ComposedNode
from ..namespace import namespace

# TODO: finish implementation to make it conformant with other ComposedNode subtypes (i.e. add merge etc.)


class ConfigTuple(ComposedNode, tuple):
    def __new__(cls, value, **kwargs):
        from .node import ConfigNode
        value = value if value is not None else tuple()
        return tuple.__new__(cls, tuple(ConfigNode(child) for child in value))

    def __init__(self, value=None, **kwargs):
        value = value if value is not None else tuple()
        #ComposedNode.maybe_inherit_flags(value, kwargs)
        kwargs.setdefault('delete', True)
        ComposedNode.__init__(self, children={ i: v for i, v in enumerate(value) }, **kwargs)

    def _validate_index(self, index):
        if not isinstance(index, int):
            raise TypeError(f'Index should be integer, got: {type(index)}')
        if abs(index) >= len(self):
            raise IndexError('Tuple index out of range')
        if index < 0:
            index = len(self) + index

        return index

    def _get(self, index, default=None, raise_ex=True):
        try:
            index = self._validate_index(index)
        except (IndexError, TypeError):
            if raise_ex:
                raise
            return default

        return tuple.__getitem__(self, index)

    def __getitem__(self, index):
        return self._get(index, raise_ex=True)

    @namespace('ayns')
    def set_child(self, index, value):
        raise TypeError('tuple does not support item assignment')

    @namespace('ayns')
    def remove_child(self, index):
        raise TypeError('tuple does not support item deletion')

    @namespace('ayns')
    def get_child(self, index, default=None):
        return self._get(index, default=default, raise_ex=False)

    @namespace('ayns')
    def map_nodes(self, map_fn):
        raw = tuple(map(map_fn, self))
        return ConfigTuple(raw)

    def __contains__(self, value):
        return tuple.__contains__(self, value)

    def __repr__(self, simple=False):
        tuple_repr = '(' + ', '.join([c.__repr__(simple=True) for c in self.ayns.children()]) + ')' # pylint: disable=no-value-for-parameter
        if simple:
            return type(self).__name__ + tuple_repr

        node = ComposedNode.__repr__(self)
        return node + tuple_repr

    def __str__(self):
        return type(self).__name__ + '(' ', '.join(map(str, self)) + ')'

    def _get_value(self):
        return self

    def _set_value(self, other):
        raise TypeError(f'Cannot set value of an immutable config node: {self!r}')



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .node import ConfigNode
from ..namespace import namespace, staticproperty


class RequiredNode(ConfigNode):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        raise ValueError(f'RequiredNode should not appear during evaluation: {path!r}')

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def tag():
        return '!required'

    @namespace('ayns')
    @property
    def value(self):
        return str()



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .node import ConfigNode, ConfigNodeMeta
from ..namespace import namespace


class configbool(int):
    def __new__(cls, value):
        return int.__new__(cls, bool(value))

    def __repr__(self):
        return bool(self).__repr__()

    def __str__(self):
        return bool(self).__str__()

    def __eq__(self, other):
        return bool(self) == other

    def __neq__(self, other):
        return bool(self) != other

    def get(self):
        return bool(self)


class ConfigNone(object):
    def __new__(cls, value):
        assert value is None
        return object.__new__(cls)

    def __repr__(self):
        return repr(None)

    def __str__(self):
        return str(None)

    def __bool__(self):
        return False

    def __eq__(self, other):
        if other is None:
            return True
        return isinstance(other, ConfigNone)

    def get(self):
        return None


class ConfigScalarMeta(ConfigNodeMeta):
    _allowed_scalar_types = { int: int, float: float, bool: configbool, str: str, type(None): ConfigNone }
    _types = {}

    def __init__(cls, name, bases, dict):
        super().__init__(name, bases, dict)
        cls._bases = bases
        cls._dict = dict

    def __call__(cls, value, **kwargs):
        if cls is not ConfigScalar:
            return super().__call__(value, **kwargs)

        type_only = False
        if isinstance(value, type):
            value_type = value
            type_only = True
        else:
            value_type = type(value)

        if not issubclass(value_type, ConfigScalarMarker):
            #if value_type not in cls._allowed_scalar_types:
            #    raise ValueError(f'Unsupported scalar type: {value_type}')
            typename = f'ConfigScalar({value_type.__name__})'
            if value_type not in cls._types:
                #bt = cls._allowed_scalar_types[value_type]
                bt = cls._allowed_scalar_types.get(value_type, value_type)
                new_value_type = ConfigScalarMeta(typename, cls._bases + (bt, ), { **cls._dict, '_dyn_base': bt } )
                cls._types[value_type] = new_value_type
                value_type = new_value_type
            else:
                value_type = cls._types[value_type]

        if type_only:
            return value_type
        ret = ConfigNodeMeta.__call__(value_type, value, **kwargs)
        ret._dyn_instance = True
        return ret

    def __instancecheck__(cls, obj):
        if cls is ConfigScalar:
            return isinstance(obj, ConfigScalarMarker)
        else:
            return super().__instancecheck__(obj)

    def __subclasscheck__(cls, subcls):
        if cls is ConfigScalar:
            return issubclass(subcls, ConfigScalarMarker)
        else:
            return super().__subclasscheck__(subcls)

    def __repr__(cls):
        clsname = '.'.join(filter(bool, [cls.__module__, cls.__name__]))
        return f'<class {clsname!r}>'


class ConfigScalarMarker(ConfigNode):
    pass


class ConfigScalar(ConfigScalarMarker, metaclass=ConfigScalarMeta):
    def __new__(cls, *value, **kwargs):
        return cls._dyn_base.__new__(cls, *value) # pylint: disable=no-member

    def __init__(self, value, **kwargs):
        ConfigNode.__init__(self, **kwargs)
        try:
            self._dyn_base.__init__(self, value, **kwargs)
        except:
            try:
                self._dyn_base.__init__(self, value)
            except:
                self._dyn_base.__init__(self) # pylint: disable=no-member

        self._dyn_instance = False

    def __repr__(self, simple=False):
        if simple:
            t_prefix = ''
            if type(self) is not ConfigScalar(self._dyn_base): # pylint: disable=no-member
                t_prefix = type(self).__name__ + '('
            return t_prefix + self._dyn_base.__repr__(self) + (')' if t_prefix else '') # pylint: disable=no-member

        node = ConfigNode.__repr__(self)
        return node + '(' + self._dyn_base.__repr__(self) + ')' # pylint: disable=no-member

    def __str__(self):
        if self._dyn_base is str:
            return str.__str__(self)

        return str(self._dyn_base(self)) # pylint: disable=no-member

    def _get_value(self):
        if self._dyn_base in [configbool, ConfigNone]: # pylint: disable=no-member
            return self._dyn_base.get(self) # pylint: disable=no-member

        return self

    def _set_value(self, other):
        raise RuntimeError(f'Cannot set value of an immutable config node: {self!r}')

    @namespace('ayns')
    def named_nodes(self, prefix='', recursive=True, include_self=True, allow_duplicates=True):
        if include_self:
            yield prefix, self

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        return self._get_native_value()

    def _get_native_value(self):
        if self._dyn_base in [configbool, ConfigNone]: # pylint: disable=no-member
            return self._get_value()

        return self._dyn_base(self) # pylint: disable=no-member

    def __reduce__(self):
        if not self._dyn_instance:
            return object.__reduce__(self)

        state = self.__dict__.copy()
        return ConfigScalar, (self._get_native_value(), ), state # pylint: disable=no-member

    def __eq__(self, other):
        if isinstance(other, ConfigScalar):
            other = other._get_native_value()
        return self._get_native_value() == other



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import pathlib

from .list import ConfigList
from .node import ConfigNode
from ..namespace import namespace, staticproperty

from collections.abc import Sequence

_parent_regexp = re.compile(r'parent(\(([0-9]+)\))?')
_abs_regexp = re.compile(r'abs\(([a-zA-Z0-9_/\ -.]+)\)')


class PathNode(ConfigList):
    ''' Implements ``!path`` tag.

        Path nodes can be used as a convenient way of defining
        paths in the underlaying file system.
        Internally, they are treated as list whose elements are
        names of components in the file system. A simple way of
        thinking about a path node is to interpret it as a list
        of arguments to ``os.path.join``.

        > Path nodes evaluate to ``pathlib.Path``.
        > Path nodes do not expand user home directory (i.e., ``~``)
        > Path nodes automatically normalize paths with ``os.path.normpath``

        In addition to holding a list of components, the path node object
        is extended with a single string specifying a *reference point*.
        This reference point defines the initial folder from which
        traversing the of file system begins.
        The end-to-end behaviour of a path node when evaluating can be
        summarized as::

            return pathlib.Path(self.get_reference_point()).joinpath(*self.components)

        Supported syntax::

            !path [names]       # long form, implicit ref. point (see below)
            !path:* [names]     # long form, explicit ref. point (see below for avail. values for *)

            !path name          # short form, eq. to: !path [name]
            !path:* name        # short form, eq. to: !path:* [name],

        Currently the following reference point specifiers are supported :

            ==============  ========================================================================
            Specifier       Value
            ==============  ========================================================================
            *(implicit)*    ``"."`` (default)
            cwd             ``os.getcwd()``
            file            name of the yaml file containing the path node
            parent          parent of the yaml file containing the path node (i.e., its folder)
            parent(n)       the n-th parent of the yaml file (0-based)
            abs(path)       user-defined ``path``
            ==============  ========================================================================


        Merge behaviour:

            The same as for list nodes.

    '''
    def __init__(self, values, ref_point, src_filename, **kwargs):
        if not isinstance(values, Sequence) or isinstance(values, str) or isinstance(values, bytes):
            if not values:
                values = []
            else:
                values = [values]

        super().__init__(values, **kwargs)

        self.ref_point = ref_point or ''
        self.src_filename = os.path.normpath(src_filename)

        self._ref_point_parsed = None
        if self.ref_point not in ['', 'cwd', 'file']:
            parent_match = _parent_regexp.match(self.ref_point)
            abs_match = _abs_regexp.match(self.ref_point)
            assert not parent_match or not abs_match
            if parent_match:
                idx = 0
                if parent_match.group(2):
                    idx = int(parent_match.group(2))
                self._ref_point_parsed = 'parent', idx
            elif abs_match:
                self._ref_point_parsed = 'abs', str(abs_match.group(1))
        else:
            self._ref_point_parsed = self.ref_point, None

        if not self._ref_point_parsed:
            raise ValueError(f'Unknown reference point provided for a PathNode: {ref_point!r}.')

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        args = ConfigList.ayns.on_evaluate(self, path, ctx)
        ref_point, ref_point_args = self._ref_point_parsed
        if ref_point == '':
            ret = pathlib.Path('.').joinpath(*args)
        elif ref_point == 'cwd':
            ret = pathlib.Path(os.getcwd()).joinpath(*args)
        elif ref_point == 'file':
            ret = pathlib.Path(self.src_filename).joinpath(*args)
        elif ref_point == 'parent':
            src = pathlib.Path(self.src_filename)
            if ref_point_args >= len(src.parents):
                diff = ref_point_args - len(src.parents) + 1
                ref_point_args = len(src.parents) - 1
                args = ['..'] * diff + args

            ret = src.parents[ref_point_args].joinpath(*args)
        elif ref_point == 'abs':
            ret = pathlib.Path(ref_point_args).joinpath(*args)
        else:
            raise ValueError(f'Unknown reference point: {ref_point!r}')

        ret = pathlib.Path(os.path.normpath(ret))
        return ret


    @namespace('ayns')
    @property
    def tag(self):
        if not self.ref_point:
            return '!path'
        return '!path:' + self.ref_point

    @namespace('ayns')
    @property
    def value(self):
        return {
            'values': list(super()._get_value()),
            'ref_point': self.ref_point,
            'src_filename': self.src_filename
        }



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dict import ConfigDict
from .function import FunctionNode
from ..namespace import namespace, staticproperty
from ..utils import import_name


class CallNode(FunctionNode):
    ''' A class implementing ``!call`` node.

        The bahaviour of this class is analogical to :py:class:`BindNode`
        but instead of evaluating to a ``functools.partial`` it calls the
        target function with provided arguments and returns its result::

            f = import_name(call_node.ayns.func)
            return f(**call_node)

        Please see the BindNode's documentation for more details.

        Supported syntax::

            !call:name { args }
            !call:name [args]    # eq. to: !call:name { enumerate(args) }
            !call name           # eq. to: !call:name {}


        Merge behaviour:

            ==================  ================================================================================================================================
            Case                Behaviour
            ==================  ================================================================================================================================
            ``dict <- Call``    behaves as ``dict <- !del dict``, unless delete is explicitly set to False
            ``None <- Call``    ``Call``
            ``Call1 <- Call2``  ``Call2`` if target entity changes, otherwise ``dict1 <- dict2``
            ``Call <- dict``    Update ``Call``'s arguments without changing target function
            ``Call <- list``    Update ``Call``'s arguments (analogical to ``dict <- list``) without changing target function
            ``Call <- str``     if ``str`` is different than the current target function's name, update the name and remove all children, otherwise no effect
            ==================  ================================================================================================================================

    '''
    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        _func = self._func
        if isinstance(_func, str):
            _func = import_name(_func)
        args = ConfigDict.ayns.on_evaluate(self, path, ctx)
        p, kw_p, kw = FunctionNode._resolve_args(_func, args)
        return _func(*p, **kw_p, **kw)

    @namespace('ayns')
    @property
    def tag(self):
        _func = self._func
        if not isinstance(_func, str):
            _func = self._func.__module__ + '.' + self._func.__name__
        return '!call:' + _func



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import config
from . import builder
from . import eval_context
from . import yaml

Config = config.Config
Builder = builder.Builder
EvalContext = eval_context.EvalContext


from .utils import add_module_properties
from .namespace import staticproperty


def _get_version():
    from . import version
    return version.version

def _get_has_repo():
    from . import version
    return version.has_repo

def _get_repo():
    from . import version
    return version.repo

def _get_commit():
    from . import version
    return version.commit


add_module_properties(__name__, {
    '__version__': staticproperty(staticmethod(_get_version)),
    '__has_repo__': staticproperty(staticmethod(_get_has_repo)),
    '__repo__': staticproperty(staticmethod(_get_repo)),
    '__commit__': staticproperty(staticmethod(_get_commit))
})



#!/usr/bin/env python

# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

import importlib.util
from pathlib import Path

package_name = 'awesomeyaml'

version_file = Path(__file__).parent.joinpath(package_name, 'version.py')
spec = importlib.util.spec_from_file_location('{}.version'.format(package_name), version_file)
package_version = importlib.util.module_from_spec(spec)
spec.loader.exec_module(package_version)

class build_maybe_inplace(build_py):
    def run(self):
        _dist_file = version_file.parent.joinpath('_dist_info.py')
        assert not _dist_file.exists()
        _dist_file.write_text('\n'.join(map(lambda attr_name: attr_name+' = '+repr(getattr(package_version, attr_name)), package_version.__all__)) + '\n')
        return super().run()

with Path(__file__).parent.joinpath('README.md').open('r') as f:
    long_desc = f.read()

setup(name=package_name,
      version=package_version.version,
      description='Config-building utilities using YAML',
      author='Łukasz Dudziak',
      author_email='l.dudziak@samsung.com',
      url='https://github.com/SamsungLabs/awesomeyaml',
      download_url='https://github.com/SamsungLabs/awesomeyaml',
      long_description=long_desc,
      long_description_content_type='text/markdown',
      python_requires='>=3.6.0',
      setup_requires=[
          'GitPython'
      ],
      install_requires=[
          'pyyaml'
      ],
      packages=find_packages(where='.', exclude=['tests']),
      package_dir={ '': '.' },
      cmdclass={
          'build_py': build_maybe_inplace
      }
)



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import threading
import contextlib
import collections.abc as cabc

from ..namespace import NamespaceableMeta, Namespace, staticproperty
from ..utils import persistent_id


class ConfigNodeMeta(NamespaceableMeta):
    def __call__(cls,
            *args,
            nodes_memo=None,
            _force_type=False,
            **kwargs):
        value = None
        has_value = False
        if args:
            value = args[0]
            args = args[1:]
            has_value = True

        # check if type deduction is required
        if cls is ConfigNode and not _force_type:
            # deduce type and call it recursively (this time enforcing it)
            if not has_value:
                raise ValueError('Cannot deduce target type without a positional argument - deduction is always done w.r.t. the first argument')
            if isinstance(value, ConfigNode):
                return value
            else:
                from .dict import ConfigDict
                from .list import ConfigList
                from .tuple import ConfigTuple
                from .scalar import ConfigScalar

                if isinstance(value, cabc.Sequence) and not isinstance(value, str) and not isinstance(value, bytes):
                    if isinstance(value, cabc.MutableSequence):
                        t = ConfigList
                    else:
                        t = ConfigTuple
                elif isinstance(value, cabc.MutableMapping):
                    t = ConfigDict
                else:
                    t = ConfigScalar

            # dispatch actual object creation (see below)
            # we need to do that recursively since __call__ method can be overwritten (e.g. ConfigScalar)
            return t(value, *args, nodes_memo=nodes_memo, _force_type=True, **kwargs)

        # actual object creation
        if has_value and nodes_memo is not None and id(value) in nodes_memo:
            return nodes_memo[id(value)]

        from .composed import ComposedNode
        if issubclass(cls, ComposedNode):
            kwargs['nodes_memo'] = nodes_memo

        if has_value:
            ret = NamespaceableMeta.__call__(cls, value, *args, **kwargs)
        else:
            assert not args
            ret = NamespaceableMeta.__call__(cls, **kwargs)

        if has_value and nodes_memo is not None:
            assert id(value) not in nodes_memo
            nodes_memo[persistent_id(value)] = ret

        return ret


class ConfigNode(metaclass=ConfigNodeMeta):
    WEAK = -1
    STANDARD = 0
    FORCE = 1

    special_metadata_names = [
        'idx',
        'merge_mode',
        'delete',
        'source_file',
        'dependencies',
        'users'
    ]

    _default_filename = threading.local()

    @staticmethod
    @contextlib.contextmanager
    def default_filename(filename):
        if not hasattr(ConfigNode._default_filename, 'value'):
            ConfigNode._default_filename.value = None

        old = ConfigNode._default_filename.value
        ConfigNode._default_filename.value = filename
        try:
            yield
        finally:
            ConfigNode._default_filename.value = old

    def __init__(self, idx=None, merge_mode=None, delete=None, metadata=None, source_file=None, implicit_delete=False):
        if merge_mode not in [None, 0, ConfigNode.WEAK, ConfigNode.FORCE]:
            raise ValueError(f'Unknown merge mode value: {merge_mode}')
        self._idx = idx
        self._merge_mode = merge_mode
        self._delete = delete
        self._implicit_delete = implicit_delete
        self._source_file = source_file if source_file is not None else getattr(ConfigNode._default_filename, 'value', None)
        self._metadata = metadata or {}

        self._default_merge_mode = ConfigNode.STANDARD
        self._default_delete = False

    def __repr__(self, simple=False):
        return f'<Object {type(self).__name__!r} at 0x{id(self):02x}>'

    class ayns(Namespace):
        @property
        def idx(self):
            return self._idx

        @property
        def merge_mode(self):
            if self._merge_mode is None:
                return self._default_merge_mode

            return self._merge_mode

        @property
        def weak(self):
            return self.ayns.merge_mode == ConfigNode.WEAK

        @property
        def force(self):
            return self.ayns.merge_mode == ConfigNode.FORCE

        @property
        def delete(self):
            if self._delete is None:
                if self._implicit_delete:
                    return True
                return self._default_delete

            return self._delete

        @property
        def explicit_delete(self):
            if self._delete is None:
                return self._default_delete
            return self._delete

        @property
        def source_file(self):
            return self._source_file

        @property
        def metadata(self):
            return self._metadata

        @staticproperty
        @staticmethod
        def is_leaf():
            return True

        @staticproperty
        @staticmethod
        def is_root():
            return False

        @property
        def value(self):
            return self._get_value()

        @property
        def native_value(self):
            return self._get_native_value()

        @value.setter
        def value(self, value):
            return self._set_value(value)

        @staticproperty
        @staticmethod
        def tag():
            return None

        def has_priority_over(self, other, if_equal=False):
            if self.ayns.merge_mode == other.ayns.merge_mode:
                return if_equal
            return self.ayns.merge_mode > other.ayns.merge_mode

        def preprocess(self, builder):
            return self.ayns.on_preprocess([], builder)

        def on_preprocess(self, path, builder):
            return self

        def premerge(self, into=None):
            return self.ayns.on_premerge([], into)

        def on_premerge(self, path, into):
            return self

        def evaluate(self):
            return self.ayns.evaluate_node([], self)

        def evaluate_node(self, path, root):
            evaluated = self.ayns.on_evaluate(path, root)
            assert evaluated is not self
            assert not isinstance(evaluated, ConfigNode)
            return evaluated

        def on_evaluate(self, path, ctx):
            return self.ayns.value

        def get_node_info_to_save(self):
            ''' This function should return a dict with values which one wants to preserve when dumping the node.
            '''
            # by default we don't care about node idx or source file (after dumping this will change anyway)
            # we care about custom metadata and "merge_mode" and "delete"
            # Note that "merge_mode" and "delete" might be optimized out from the dump output if
            # they would be set by the parent (dumping function holds a stack of which metadata are "default"
            # and does not produce anything which is aligned with the defaults)
            ret = copy.copy(self._metadata)
            ret['merge_mode'] = self._merge_mode
            ret['delete'] = self._delete #if not self._implicit_delete else None
            return ret

        @property
        def node_info(self):
            return {
                'idx': self._idx,
                'merge_mode': self._merge_mode,
                'delete': self._delete,
                'implicit_delete': self._implicit_delete,
                'source_file': self._source_file,
                'metadata': self._metadata
            }

        def get_default_mode(self):
            return {
                'merge_mode': self._default_merge_mode,
                'delete': self._default_delete
            }

        def represent(self):
            ''' Returns a tuple ``(tag, metadata, data)``, where ``tag`` is desired tag (can be ``None``),
                ``metadata`` is a dict with metadata (optional, can evaluate to ``False`` to ignore),
                and ``data`` is object which will be used to recursively represent ``self`` (can be either
                mapping, sequence or scalar).
            '''
            return self.ayns.tag, self.ayns.get_node_info_to_save(), self.ayns.value



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .scalar import ConfigScalar
from .node import ConfigNode
from ..namespace import namespace, staticproperty
from ..errors import EvalError
from ..utils import Bunch

import os
import sys
import copy
import types
import ctypes
import hashlib


class EvalNode(ConfigScalar(str)):
    ''' Implements ``!eval`` tag.

        The eval node can be used to evaluate arbitrary python code.
        Internally, it is treated exactly as normal string node with the
        only exception being that on evaluation its value is passed to ``eval``.

        More specifically, the value obtained from evaluating an eval node is the
        value returned by the last line of code store in the node. If more lines are
        present, the preceding ones are executed before the last one to provide
        additional context (e.g., by defining functions, importing modules etc.).

        On top of that, the code stored in an eval node can use symbols defined
        by a relevant :py:class:`EvalContext` (please see its constructor)
        and can access all top-level nodes in the config by their names.

        The summary of the eval node's behaviour is illustrated by the following snippet::

            symbols = eval_context.get_eval_symbols()
            symbols.update(eval_context.get_top_level_objects_and_names())
            lines = str(self.code).split('\\n')
            exec('\\n'.join(lines[:-1]), symbols)
            return eval(lines[-1], symbols)

        Supported syntax::

            !eval code

        Merge behaviour:

            The same as standard string node.
    '''
    _top_namespace_module_name = 'awesomeyaml.eval_node_namespace'

    def __init__(self, value, persistent_namespace=True, source_file=None, **kwargs):
        super().__init__(value, **kwargs)
        self.persistent_namespace = persistent_namespace

    @namespace('ayns')
    def on_evaluate(self, path, ctx):
        code_hash = hashlib.md5(str(self).encode('utf-8')).hexdigest()
        eval_module_name = f'{EvalNode._top_namespace_module_name}.{str(path).replace(".", "_")}_0x{code_hash}'

        gbls = { 'ayns': Bunch({
                'ctx': ctx
            })
        }
        gbls.update(copy.copy(ctx.get_eval_symbols()))
        gbls.update(dict(ctx.ayns.named_children()))
        gbls.update({ '__name__': eval_module_name, '__file__': self._source_file })

        lines = self.strip().split('\n')
        lines = [lline for line in lines for lline in line.split(';')]
        try:
            exec("\n".join(lines[:-1]), gbls)
            ret = eval(lines[-1].strip(), gbls)
        except:
            et, e, _ = sys.exc_info()
            raise EvalError(f'Exception occurred while evaluation an eval node {path!r} from file {str(self.ayns.source_file)!r}:\n\nCode:\n{os.linesep.join(lines)}\n\nError:\n{et.__name__}: {e}') from None

        if len(lines) > 1 and self.persistent_namespace:
            eval_node_module = types.ModuleType(eval_module_name, 'Dynamic module to evaluate awesomeyaml !eval node')
            eval_node_module.__dict__.update(gbls)
            sys.modules[eval_module_name] = eval_node_module

        if isinstance(ret, ConfigNode):
            assert ret is not self
            ret = ret.ayns.on_evaluate(path, ctx)
        return ret

    @namespace('ayns')
    @staticproperty
    @staticmethod
    def tag():
        return '!eval'



# Copyright 2022 Samsung Electronics Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
import re
import token
import pickle
import tokenize
import contextlib
import collections.abc as cabc

from .nodes.node import ConfigNode
from .nodes.composed import ComposedNode
from .utils import pad_with_none


_fstr_regex = re.compile(r"^\s*f(['\"]).*\1\s*$")

_global_ctx = None


@contextlib.contextmanager
def global_ctx(filename):
    global _global_ctx
    if _global_ctx is None:
        import awesomeyaml.builder as b
        _global_ctx = b.Builder()

    _global_ctx._current_file = filename
    with ConfigNode.default_filename(filename):
        yield _global_ctx
    _global_ctx._current_file = None



def _encode_metadata(metadata):
    return pickle.dumps(metadata).hex()


def _decode_metadata(encoded):
    if not encoded:
        return None
    return pickle.loads(bytes.fromhex(encoded))


def _maybe_parse_scalar(loader, node, reparse=True):
    ret = loader.construct_scalar(node)
    if reparse and isinstance(ret, str) and not ret.strip().startswith('!'):
        ret = yaml.safe_load(ret)

    return ret


def _make_obj(loader, node, objtype, reparse_scalars=False):
    args = []
    kwargs = {}
    if isinstance(node, yaml.MappingNode):
        kwargs = loader.construct_mapping(node, deep=True)
    elif isinstance(node, yaml.SequenceNode):
        args = loader.construct_sequence(node, deep=True)
    elif isinstance(node, yaml.ScalarNode):
        val = _maybe_parse_scalar(loader, node, reparse=reparse_scalars)
        if val:
            args = [val]

    if '*' in kwargs:
        args.extend(kwargs['*'])
        del kwargs['*']

    return objtype(*args, **kwargs)


def _make_node(loader, node, node_type=ConfigNode, kwargs=None, data_arg_name=None, dict_is_data=True):
    ''' A generic function to create new config nodes.

        Arguments:
            loader : a yaml.Loader
            node : a yaml node to parse, which will be used to create a config node
            node_type : a callable which will be called to create a config node - the following arguments control
                what are the arguments to the callable
            kwargs : a fixed directory of extra keyword arguments which will be passed to ``node_type``
            data_arg_name : if parsed node data should be passed as a keyword argument, the name of the
                argument should be specified by this argument, if it is ``None``, parsed node data
                will be passed as the first (positional) argument::

                    data = parse(node)
                    if data_arg_name:
                        kwargs[data_arg_name] = data
                        return node_type(**kwargs)
                    else:
                        return node_type(data, **kwargs)

            dict_is_data : if ``True`` and the parsed node data is mapping, the data will be used as
                a single ``"data"`` argument (see ``data_arg_name``), otherwise if ``False`` and the parsed
                node data is mapping, the dict will be treated as ``**kwargs`` for ``node_type``::

                    data = parse(node)
                    if isinstance(data, dict) and not dict_is_data:
                        kwargs.update(data)
                        return node_type(**kwargs)
                    else:
                        return node_type(data, **kwargs)

    '''
    kwargs = kwargs or {}
    data = None
    is_dict = False
    if isinstance(node, yaml.MappingNode):
        data = loader.construct_mapping(node, deep=True)
        is_dict = True
    elif isinstance(node, yaml.SequenceNode):
        data = loader.construct_sequence(node, deep=True)
    elif isinstance(node, yaml.ScalarNode):
        data = _maybe_parse_scalar(loader, node, reparse=node_type is ConfigNode)

    kwargs.setdefault('source_file', loader.context.get_current_file())

    if is_dict and not dict_is_data:
        kwargs.update(data)
        return node_type(**kwargs)

    if data_arg_name is None:
        return node_type(data, idx=loader.context.get_next_stage_idx(), **kwargs)
    else:
        assert data_arg_name not in kwargs
        kwargs[data_arg_name] = data
        return node_type(idx=loader.context.get_next_stage_idx(), **kwargs)


def _del_constructor(loader, node):
    return _make_node(loader, node, kwargs={ 'delete': True })


def _weak_constructor(loader, node):
    return _make_node(loader, node, kwargs={ 'merge_mode': ConfigNode.WEAK })


def _force_constructor(loader, node):
    return _make_node(loader, node, kwargs={ 'merge_mode': ConfigNode.FORCE })


def _merge_constructor(loader, node):
    return _make_node(loader, node, kwargs={ 'delete': False })


def _append_constructor(loader, node):
    from .nodes.append import AppendNode
    return _make_node(loader, node, node_type=AppendNode)


def _metadata_constructor(loader, tag_suffix, node):
    metadata = _decode_metadata(tag_suffix)
    kwargs = {}
    for special in ConfigNode.special_metadata_names:
        if special in metadata:
            kwargs[special] = metadata.pop(special)

    kwargs['metadata'] = metadata
    return _make_node(loader, node, kwargs=kwargs)


def _include_constructor(loader, node):
    from .nodes.include import IncludeNode
    return _make_node(loader, node, node_type=IncludeNode, dict_is_data=False)


def _prev_constructor(loader, node):
    from .nodes.prev import PrevNode
    return _make_node(loader, node, node_type=PrevNode)


def _xref_constructor(loader, node):
    from .nodes.xref import XRefNode
    return _make_node(loader, node, node_type=XRefNode)

def _xref_constructor_md(loader, tag_suffix, node):
    metadata = _decode_metadata(tag_suffix)
    kwargs = {}
    for special in ConfigNode.special_metadata_names:
        if special in metadata:
            kwargs[special] = metadata.pop(special)

    kwargs['metadata'] = metadata

    from .nodes.xref import XRefNode
    return _make_node(loader, node, kwargs=kwargs, node_type=XRefNode)


def _simple_bind_constructor(loader, node):
    from .nodes.bind import BindNode
    return _make_node(loader, node, node_type=BindNode, data_arg_name='func')


def _bind_constructor(loader, tag_suffix, node):
    from .nodes.bind import BindNode
    if tag_suffix.count(':') > 1:
        raise ValueError(f'Invalid bind tag: !bind:{tag_suffix}')

    target_f_name, metadata = pad_with_none(*tag_suffix.split(':', maxsplit=1), minlen=2)
    metadata = _decode_metadata(metadata)
    return _make_node(loader, node, node_type=BindNode, kwargs={ 'func': target_f_name, 'metadata': metadata }, data_arg_name='args')


def _simple_call_constructor(loader, node):
    from .nodes.call import CallNode
    return _make_node(loader, node, node_type=CallNode, data_arg_name='func')


def _call_constructor(loader, tag_suffix, node):
    from .nodes.call import CallNode
    if tag_suffix.count(':') > 1:
        raise ValueError(f'Invalid call tag: !call:{tag_suffix}')

    target_f_name, metadata = pad_with_none(*tag_suffix.split(':', maxsplit=1), minlen=2)
    metadata = _decode_metadata(metadata)
    return _make_node(loader, node, node_type=CallNode, kwargs={ 'func': target_f_name, 'metadata': metadata }, data_arg_name='args')



def _eval_constructor(loader, tag_suffix, node):
    from .nodes.eval import EvalNode
    metadata = _decode_metadata(tag_suffix)
    kwargs = {}
    for special in ConfigNode.special_metadata_names:
        if special in metadata:
            kwargs[special] = metadata.pop(special)

    kwargs['metadata'] = metadata
    return _make_node(loader, node, node_type=EvalNode, kwargs=kwargs)


def _simple_eval_constructor(loader, node):
    from .nodes.eval import EvalNode
    return _make_node(loader, node, node_type=EvalNode)


def _fstr_constructor(loader, node):
    from .nodes.fstr import FStrNode

    def _maybe_fix_fstr(value, *args, **kwargs):
        try:
            return FStrNode(value, *args, **kwargs)
        except ValueError:
            return FStrNode("f'" + value.replace(r"'", r"\'") + "'", *args, **kwargs)

    return _make_node(loader, node, node_type=_maybe_fix_fstr)


def _import_constructor(loader, node):
    import importlib
    module = importlib.import_module('.nodes.import', package='awesomeyaml') # dirty hack because "import" is a keyword
    ImportNode = module.ImportNode
    return _make_node(loader, node, node_type=ImportNode)


def _required_constructor(loader, node):
    from .nodes.required import RequiredNode
    def _check_empty_str(arg, **kwargs):
        if arg != '':
            raise ValueError(f'!required node does not expect any arguments - got: {arg}')
        return RequiredNode(**kwargs)
    return _make_node(loader, node, node_type=_check_empty_str)


def _required_constructor_md(loader, tag_suffix, node):
    metadata = _decode_metadata(tag_suffix)
    kwargs = {}
    for special in ConfigNode.special_metadata_names:
        if special in metadata:
            kwargs[special] = metadata.pop(special)

    kwargs['metadata'] = metadata

    from .nodes.required import RequiredNode
    def _check_empty_str(arg, **kwargs):
        if arg != '':
            raise ValueError(f'!required node does not expect any arguments - got: {arg}')
        return RequiredNode(**kwargs)
    return _make_node(loader, node, kwargs=kwargs, node_type=_check_empty_str)


def _none_constructor(loader, node):
    def _check_empty_str(arg, *args, **kwargs):
        if arg != '' or args:
            raise ValueError(f'!null does not expect any arguments - got: {[arg]+list(args)}')
        return None
    return _make_node(loader, node, node_type=_check_empty_str)


def _none_constructor_md(loader, tag_suffix, node):
    metadata = _decode_metadata(tag_suffix)
    kwargs = {}
    for special in ConfigNode.special_metadata_names:
        if special in metadata:
            kwargs[special] = metadata.pop(special)

    kwargs['metadata'] = metadata

    def _check_empty_str(arg, *args, **kwargs):
        if arg != '' or args:
            raise ValueError(f'!null does not expect any arguments - got: {[arg]+list(args)}')
        return None
    return _make_node(loader, node, kwargs=kwargs, node_type=_check_empty_str)


def _simple_path_constructor(loader, node):
    from .nodes.path import PathNode
    return _make_node(loader, node, node_type=PathNode, kwargs={ 'ref_point': None, 'src_filename': loader.context.get_current_file() })


def _path_constructor(loader, tag_suffix, node):
    from .nodes.path import PathNode
    if tag_suffix.count(':') > 1:
        raise ValueError(f'Invalid path tag: !path:{tag_suffix}')

    ref_point, metadata = pad_with_none(*tag_suffix.rsplit(':', maxsplit=1), minlen=2)
    metadata = _decode_metadata(metadata)
    return _make_node(loader, node, node_type=PathNode, kwargs={ 'ref_point': ref_point, 'src_filename': loader.context.get_current_file(), 'metadata': metadata }, dict_is_data=False)


def make_call_node_with_fixed_func(loader, node, func):
    from .nodes.call import CallNode
    return _make_node(loader, node, node_type=CallNode, kwargs={ 'func': func }, data_arg_name='args')


yaml.add_constructor('!del', _del_constructor)
yaml.add_constructor('!weak', _weak_constructor)
yaml.add_constructor('!force', _force_constructor)
yaml.add_constructor('!merge', _merge_constructor)
yaml.add_constructor('!append', _append_constructor)
yaml.add_multi_constructor('!metadata:', _metadata_constructor)
yaml.add_constructor('!include', _include_constructor)
yaml.add_constructor('!prev', _prev_constructor)
yaml.add_constructor('!xref', _xref_constructor)
yaml.add_multi_constructor('!xref:', _xref_constructor_md)
yaml.add_constructor('!ref', _xref_constructor)
yaml.add_multi_constructor('!ref:', _xref_constructor_md)
yaml.add_multi_constructor('!bind:', _bind_constructor) # full bind form: !bind:func_name[:metadata] args_dict
yaml.add_constructor('!bind', _simple_bind_constructor) # simple argumentless bind from string: !bind func_name
yaml.add_multi_constructor('!call:', _call_constructor) # full call form: !call:func_name[:metadata] args_dict
yaml.add_constructor('!call', _simple_call_constructor) # simple argumentless call from string: !call func_name
yaml.add_multi_constructor('!eval:', _eval_constructor)
yaml.add_constructor('!eval', _simple_eval_constructor)
yaml.add_constructor('!fstr', _fstr_constructor)
yaml.add_implicit_resolver('!fstr', _fstr_regex)
yaml.add_constructor('!import', _import_constructor)
yaml.add_constructor('!required', _required_constructor)
yaml.add_multi_constructor('!required:', _required_constructor_md)
yaml.add_constructor('!null', _none_constructor)
yaml.add_multi_constructor('!null:', _none_constructor_md)
yaml.add_multi_constructor('!path:', _path_constructor)
yaml.add_constructor('!path', _simple_path_constructor)


def _node_representer(dumper, node):
    from .nodes.bind import BindNode
    tag, metadata, data = node.ayns.represent()
    if data is None:
        assert not tag
        tag = '!null'

    parent_metadata = dumper.metadata[-1] if dumper.metadata else {}
    type_defaults = node.ayns.get_default_mode()

    tags_to_infer = {
        'merge_mode': {
            ConfigNode.STANDARD: '',
            ConfigNode.WEAK: '!weak',
            ConfigNode.FORCE: '!force'
        },
        'delete': {
            True: '!del',
            False: '!merge'
        }
    }

    to_infer = list(tags_to_infer.keys())

    for f in to_infer:
        if f not in metadata:
            continue

        current = metadata[f]
        parent = parent_metadata.get(f, None) if parent_metadata else None
        default = type_defaults[f]
        if current is not None:
            if current == parent or current == default:
                del metadata[f]
        else:
            del metadata[f]

    metadata = { key: value for key, value in metadata.items() if key not in dumper.exclude_metadata }

    # try to use simple standard tag rather then encoded metadata
    # this is possible if we only have one special thing to handle
    # (e.g. delete is set to True)
    # if more then one things are changed for a particular node, we need
    # to se !metadata:hash anyway since it's impossible to have more
    # then two tags at the same time
    if not tag and len(metadata) == 1:
        # check if the only element in metadata is one of the standard
        # things which can be controller with simple tags (those listed
        # in "tags_to_infer")
        key = next(iter(metadata.keys()))
        maybe_tag = tags_to_infer.get(key)
        if maybe_tag:
            tag = maybe_tag[metadata[key]]
            del metadata[key]

    if metadata:
        if tag is None:
            tag = '!metadata'

        tag += ':' + _encode_metadata(metadata)

    pop = False
    if isinstance(node, ComposedNode):
        dumper.metadata.append({ **parent_metadata, **metadata })
        pop = True

    try:
        if not tag and not isinstance(data, ConfigNode):
            return dumper.represent_data(data)

        if isinstance(data, cabc.Mapping):
            if tag:
                return dumper.represent_mapping(tag, data)
            else:
                data = dict(data)
                return dumper.represent_data(data)

        elif isinstance(data, cabc.Sequence) and not isinstance(data, str) and not isinstance(data, bytes):
            if tag:
                return dumper.represent_sequence(tag, data)
            else:
                if isinstance(data, cabc.MutableSequence):
                    data = list(data)
                else:
                    data = tuple(data)
                return dumper.represent_data(data)
        else:
            if tag:
                if data is None:
                    assert tag.startswith('!null')
                    return dumper.represent_scalar(tag, str(''))
                return dumper.represent_scalar(tag, str(data))
            else:
                from .nodes.scalar import ConfigScalar
                if isinstance(data, ConfigScalar):
                    return dumper.represent_data(data._dyn_base(data))
                else:
                    # fallback to str
                    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))
    finally:
        if pop:
            dumper.metadata.pop()


def _none_representer(dumper, none):
    return dumper.dump_scalar('!null', str(''))


yaml.add_multi_representer(ConfigNode, _node_representer)
yaml.add_representer(type(None), _none_representer)

def _get_metadata_end(data, beg):
    _beg = beg+2
    def _readline():
        nonlocal _beg
        end = data.find('}}', _beg)
        if end == -1:
            end = len(data)
        else:
            end += 2

        ret = data[_beg:end]
        _beg = end
        return ret.encode('utf8')
         
    last_close = False
    end = None
    for tok in tokenize.tokenize(_readline):
        if tok.type == token.OP and tok.string == '}':
            if last_close:
                end = _beg
                break
            else:
                last_close = True
        else:
            last_close = False

    return end


def _get_metadata_content(data):
    _metadata_tag = re.compile(r'(![a-zA-Z0-9_:.]+){{')
    curr_match = _metadata_tag.search(data)
    while curr_match is not None:
        beg = curr_match.end(1)
        assert data[beg:beg+2] == '{{'
        end = _get_metadata_end(data, beg)
        if end is None:
            raise ValueError(f'Cannot find the end of a !metadata node which begins at: {curr_match.start()}')
        
        yield beg, end
        curr_match = _metadata_tag.search(data, end+1)


def _encode_all_metadata(data):
    ranges = list(_get_metadata_content(data))
    offset = 0
    for beg, end in ranges:
        beg += offset
        end += offset

        metadata = eval(data[beg+1:end-1])
        encoded = _encode_metadata(metadata)
        repl = ':' + encoded
        orig_len = end-beg
        repl_len = len(repl)
        data = data[:beg] + repl + data[end:]
        offset += repl_len - orig_len

    return data


def parse(data, filename_or_builder=None, config_nodes=True):
    if not isinstance(data, str):
        data = data.read()

    @contextlib.contextmanager
    def _dummy(context):
        yield context

    try:
        filename_or_builder.get_current_stage_idx()
    except:
        # filename_or_builder doesn't seem to be a builder
        # use global context
        context_fn = global_ctx
    else:
        # filename_or_builder behaves like builder so let's use it
        # as it is
        context_fn = _dummy

    #print(data)
    with context_fn(filename_or_builder) as context:
        data = _encode_all_metadata(data)
        def get_loader(*args, **kwargs):
            loader = yaml.Loader(*args, **kwargs)
            loader.context = context
            loader.name = context.get_current_file()
            return loader

        for raw in yaml.load_all(data, Loader=get_loader):
            if config_nodes:
                yield ConfigNode(raw)
            else:
                yield ConfigNode(raw).ayns.native_value


def dump(nodes, output=None, open_mode='w', exclude_metadata=None, **kwargs):
    close = False
    if isinstance(output, str):
        output = open(output, open_mode)
        close = True

    def get_dumper(*args, **kwargs):
        dumper = yaml.Dumper(*args, **kwargs)
        assert not hasattr(dumper, 'metadata')
        dumper.metadata = []
        dumper.exclude_metadata = exclude_metadata or set()
        return dumper

    try:
        ret = yaml.dump(ConfigNode(nodes), stream=output, Dumper=get_dumper, **kwargs)
    finally:
        if close:
            output.close()

    if output is None:
        return ret
    return None
